# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ...config import options
from ...core import OutputType
from ...serialize import AnyField
from .core import DataFrameReductionOperand, DataFrameReductionMixin


class DataFrameCustomReduction(DataFrameReductionOperand, DataFrameReductionMixin):
    _op_type_ = OperandDef.CUSTOM_REDUCTION
    _func_name = 'custom_reduction'

    _custom_reduction = AnyField('custom_reduction')

    def __init__(self, custom_reduction=None, **kw):
        super().__init__(_custom_reduction=custom_reduction, **kw)

    @property
    def custom_reduction(self):
        return self._custom_reduction

    @property
    def is_atomic(self):
        return True

    def get_reduction_args(self, axis=None):
        return dict()


def build_custom_reduction_result(df, custom_reduction_obj):
    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    output_type = OutputType.series if df.ndim == 2 else OutputType.scalar
    op = DataFrameCustomReduction(custom_reduction=custom_reduction_obj, output_types=[output_type],
                                  use_inf_as_na=use_inf_as_na)
    return op(df)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes
from ...custom_log import redirect_custom_log
from ...serialize import KeyField, FunctionField, TupleField, \
    DictField, StringField
from ...tiles import TilesError
from ...utils import check_chunks_unknown_shape, enter_current_session, quiet_stdio
from ..operands import DataFrameOperand, DataFrameOperandMixin, OutputType
from ..utils import build_df, build_series, build_empty_df, parse_index, \
    validate_output_types


class DataFrameCartesianChunk(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.CARTESIAN_CHUNK

    _left = KeyField('left')
    _right = KeyField('right')
    _func = FunctionField('func')
    _args = TupleField('args')
    _kwargs = DictField('kwargs')
    # for chunk
    _tileable_op_key = StringField('tileable_op_key')

    def __init__(self, left=None, right=None, func=None, args=None, kwargs=None,
                 memory_scale=None, output_types=None, tileable_op_key=None, **kw):
        super().__init__(_left=left, _right=right, _func=func, _args=args,
                         _kwargs=kwargs, _memory_scale=memory_scale,
                         _output_types=output_types,
                         _tileable_op_key=tileable_op_key, **kw)

    @property
    def left(self):
        return self._left

    @property
    def right(self):
        return self._right

    @property
    def func(self):
        return self._func

    @property
    def args(self):
        return self._args

    @property
    def kwargs(self):
        return self._kwargs

    @property
    def memory_scale(self):
        return self._memory_scale or 2.0

    @property
    def tileable_op_key(self):
        return self._tileable_op_key

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._left = self._inputs[0]
        self._right = self._inputs[1]

    @staticmethod
    def _build_test_obj(obj):
        return build_df(obj, size=2) if obj.ndim == 2 else \
            build_series(obj, size=2, name=obj.name)

    def __call__(self, left, right, index=None, dtypes=None):
        test_left = self._build_test_obj(left)
        test_right = self._build_test_obj(right)
        output_type = self._output_types[0] if self._output_types else None

        # try run to infer meta
        try:
            with np.errstate(all='ignore'), quiet_stdio():
                obj = self._func(test_left, test_right, *self._args, **self._kwargs)
        except:  # noqa: E722  # nosec  # pylint: disable=bare-except
            if output_type == OutputType.series:
                obj = pd.Series([], dtype=np.dtype(object))
            elif output_type == OutputType.dataframe and dtypes is not None:
                obj = build_empty_df(dtypes)
            else:
                raise TypeError('Cannot determine `output_type`, '
                                'you have to specify it as `dataframe` or `series`, '
                                'for dataframe, `dtypes` is required as well '
                                'if output_type=\'dataframe\'')

        if getattr(obj, 'ndim', 0) == 1 or output_type == OutputType.series:
            shape = self._kwargs.pop('shape', (np.nan,))
            if index is None:
                index = obj.index
            index_value = parse_index(index, left, right,
                                      self._func, self._args, self._kwargs)
            return self.new_series([left, right], dtype=obj.dtype,
                                   shape=shape, index_value=index_value,
                                   name=obj.name)
        else:
            dtypes = dtypes if dtypes is not None else obj.dtypes
            # dataframe
            shape = (np.nan, len(dtypes))
            columns_value = parse_index(dtypes.index, store_data=True)
            if index is None:
                index = obj.index
            index_value = parse_index(index, left, right,
                                      self._func, self._args, self._kwargs)
            return self.new_dataframe([left, right], shape=shape,
                                      dtypes=dtypes, index_value=index_value,
                                      columns_value=columns_value)

    @classmethod
    def tile(cls, op: "DataFrameCartesianChunk"):
        left = op.left
        right = op.right
        out = op.outputs[0]

        if left.ndim == 2 and left.chunk_shape[1] > 1:
            check_chunks_unknown_shape([left], TilesError)
            # if left is a DataFrame, make sure 1 chunk on axis columns
            left = left.rechunk({1: left.shape[1]})._inplace_tile()
        if right.ndim == 2 and right.chunk_shape[1] > 1:
            check_chunks_unknown_shape([right], TilesError)
            # if right is a DataFrame, make sure 1 chunk on axis columns
            right = right.rechunk({1: right.shape[1]})._inplace_tile()

        out_chunks = []
        nsplits = [[]] if out.ndim == 1 else [[], [out.shape[1]]]
        i = 0
        for left_chunk in left.chunks:
            for right_chunk in right.chunks:
                chunk_op = op.copy().reset_key()
                chunk_op._tileable_op_key = op.key
                if op.output_types[0] == OutputType.dataframe:
                    shape = (np.nan, out.shape[1])
                    index_value = parse_index(out.index_value.to_pandas(),
                                              left_chunk, right_chunk,
                                              op.func, op.args, op.kwargs)
                    out_chunk = chunk_op.new_chunk([left_chunk, right_chunk],
                                                   shape=shape,
                                                   index_value=index_value,
                                                   columns_value=out.columns_value,
                                                   dtypes=out.dtypes,
                                                   index=(i, 0))
                    out_chunks.append(out_chunk)
                    nsplits[0].append(out_chunk.shape[0])
                else:
                    shape = (np.nan,)
                    index_value = parse_index(out.index_value.to_pandas(),
                                              left_chunk, right_chunk,
                                              op.func, op.args, op.kwargs)
                    out_chunk = chunk_op.new_chunk([left_chunk, right_chunk],
                                                   shape=shape,
                                                   index_value=index_value,
                                                   dtype=out.dtype,
                                                   name=out.name,
                                                   index=(i,))
                    out_chunks.append(out_chunk)
                    nsplits[0].append(out_chunk.shape[0])
                i += 1

        params = out.params
        params['nsplits'] = tuple(tuple(ns) for ns in nsplits)
        params['chunks'] = out_chunks
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    @redirect_custom_log
    @enter_current_session
    def execute(cls, ctx, op: "DataFrameCartesianChunk"):
        left, right = ctx[op.left.key], ctx[op.right.key]
        ctx[op.outputs[0].key] = op.func(
            left, right, *op.args, **(op.kwargs or dict()))


def cartesian_chunk(left, right, func, args=(), **kwargs):
    output_type = kwargs.pop('output_type', None)
    output_types = kwargs.pop('output_types', None)
    object_type = kwargs.pop('object_type', None)
    output_types = validate_output_types(
        output_type=output_type, output_types=output_types, object_type=object_type)
    output_type = output_types[0] if output_types else None
    if output_type:
        output_types = [output_type]
    index = kwargs.pop('index', None)
    dtypes = kwargs.pop('dtypes', None)
    memory_scale = kwargs.pop('memory_scale', None)

    op = DataFrameCartesianChunk(left=left, right=right, func=func,
                                 args=args, kwargs=kwargs,
                                 output_types=output_types,
                                 memory_scale=memory_scale)
    return op(left, right, index=index, dtypes=dtypes)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from typing import Tuple

import numpy as np

from .... import opcodes as OperandDef
from ....serialize import KeyField, AnyField, Float16Field
from ....utils import check_chunks_unknown_shape, require_module
from ....tiles import TilesError
from ...operands import TensorOperand, TensorOperandMixin
from ...core import TensorOrder
from ...datasource import tensor as astensor
from ...array_utils import as_same_device, cp, device


class TensorCdist(TensorOperand, TensorOperandMixin):
    _op_type_ = OperandDef.CDIST

    _xa = KeyField('XA')
    _xb = KeyField('XB')
    _metric = AnyField('metric')
    _p = Float16Field('p', on_serialize=lambda x: float(x) if x is not None else x)
    _w = KeyField('w')
    _v = KeyField('V')
    _vi = KeyField('VI')

    def __init__(self, metric=None, p=None, w=None,
                 v=None, vi=None, dtype=None, **kw):
        super().__init__(_metric=metric, _p=p,
                         _w=w, _v=v, _vi=vi, _dtype=dtype, **kw)

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        self._xa = next(inputs_iter)
        self._xb = next(inputs_iter)
        if self._w is not None:
            self._w = next(inputs_iter)
        if self._v is not None:
            self._v = next(inputs_iter)
        if self._vi is not None:
            self._vi = next(inputs_iter)

    @property
    def xa(self):
        return self._xa

    @property
    def xb(self):
        return self._xb

    @property
    def metric(self):
        return self._metric

    @property
    def p(self):
        return self._p

    @property
    def w(self):
        return self._w

    @property
    def v(self):
        return self._v

    @property
    def vi(self):
        return self._vi

    def __call__(self, xa, xb, shape: Tuple):
        inputs = [xa, xb]
        for val in [self._w, self._v, self._vi]:
            if val is not None:
                inputs.append(val)
        return self.new_tensor(inputs, shape=shape, order=TensorOrder.C_ORDER)

    @classmethod
    def _tile_one_chunk(cls, op, xa, xb, w, v, vi):
        out_tensor = op.outputs[0]
        chunk_op = op.copy().reset_key()
        chunk_inputs = [xa.chunks[0], xb.chunks[0]]
        for val in [w, v, vi]:
            if val is not None:
                chunk_inputs.append(val.chunks[0])
        chunk = chunk_op.new_chunk(chunk_inputs, shape=out_tensor.shape,
                                   order=out_tensor.order, index=(0,) * out_tensor.ndim)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, shape=out_tensor.shape,
                                  order=out_tensor.order,
                                  nsplits=tuple((s,) for s in out_tensor.shape),
                                  chunks=[chunk])

    @classmethod
    def _tile_chunks(cls, op, xa, xb, w, v, vi):
        out_tensor = op.outputs[0]
        acs, bcs = xa.chunk_shape[0], xb.chunk_shape[0]

        out_chunks = []
        for idx in itertools.product(range(acs), range(bcs)):
            ixa, ixb = idx
            chunk_op = op.copy().reset_key()

            chunk_inputs = []
            xa_chunk = xa.cix[ixa, 0]
            xb_chunk = xb.cix[ixb, 0]
            chunk_inputs.extend([xa_chunk, xb_chunk])
            if w is not None:
                w_chunk = chunk_op._w = w.chunks[0]
                chunk_inputs.append(w_chunk)
            if v is not None:
                v_chunk = chunk_op._v = v.chunks[0]
                chunk_inputs.append(v_chunk)
            if vi is not None:
                vi_chunk = chunk_op._vi = vi.chunks[0]
                chunk_inputs.append(vi_chunk)
            chunk = chunk_op.new_chunk(
                chunk_inputs, shape=(xa_chunk.shape[0], xb_chunk.shape[0]),
                order=out_tensor.order, index=idx)
            out_chunks.append(chunk)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, shape=out_tensor.shape,
                                  order=out_tensor.order,
                                  chunks=out_chunks,
                                  nsplits=(xa.nsplits[0], xb.nsplits[0]))

    @classmethod
    def tile(cls, op):
        # make sure every inputs have known shape
        check_chunks_unknown_shape(op.inputs, TilesError)

        xa = op.xa.rechunk({1: op.xa.shape[1]})._inplace_tile()
        xb = op.xb.rechunk({1: op.xb.shape[1]})._inplace_tile()

        # rechunk w, v, vi into one chunk if any of them has value
        extra_inputs = [None] * 3
        for i, ei in enumerate([op.w, op.v, op.vi]):
            if ei is None:
                continue
            new_ei = ei.rechunk(ei.shape)._inplace_tile()
            extra_inputs[i] = new_ei
        w, v, vi = extra_inputs

        if len(xa.chunks) == 1 and len(xb.chunks) == 1:
            # only 1 chunk
            return cls._tile_one_chunk(op, xa, xb, w, v, vi)
        else:
            return cls._tile_chunks(op, xa, xb, w, v, vi)

    @classmethod
    def execute(cls, ctx, op):
        from scipy.spatial.distance import cdist

        inputs, device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs], device=op.device, ret_extra=True)

        if xp is cp:  # pragma: no cover
            raise NotImplementedError('`cdist` does not support running on GPU yet')

        with device(device_id):
            inputs_iter = iter(inputs)
            xa = next(inputs_iter)
            xb = next(inputs_iter)
            kw = dict()
            if op.p is not None:
                kw['p'] = op.p
            if op.w is not None:
                kw['w'] = next(inputs_iter)
            if op.v is not None:
                kw['V'] = next(inputs_iter)
            if op.vi is not None:
                kw['VI'] = next(inputs_iter)

        ctx[op.outputs[0].key] = cdist(xa, xb, metric=op.metric, **kw)


@require_module('scipy.spatial.distance')
def cdist(XA, XB, metric='euclidean', **kwargs):
    """
    Compute distance between each pair of the two collections of inputs.

    See Notes for common calling conventions.

    Parameters
    ----------
    XA : Tensor
        An :math:`m_A` by :math:`n` tensor of :math:`m_A`
        original observations in an :math:`n`-dimensional space.
        Inputs are converted to float type.
    XB : Tensor
        An :math:`m_B` by :math:`n` tensor of :math:`m_B`
        original observations in an :math:`n`-dimensional space.
        Inputs are converted to float type.
    metric : str or callable, optional
        The distance metric to use.  If a string, the distance function can be
        'braycurtis', 'canberra', 'chebyshev', 'cityblock', 'correlation',
        'cosine', 'dice', 'euclidean', 'hamming', 'jaccard', 'jensenshannon',
        'kulsinski', 'mahalanobis', 'matching', 'minkowski', 'rogerstanimoto',
        'russellrao', 'seuclidean', 'sokalmichener', 'sokalsneath', 'sqeuclidean',
        'wminkowski', 'yule'.
    **kwargs : dict, optional
        Extra arguments to `metric`: refer to each metric documentation for a
        list of all possible arguments.

        Some possible arguments:

        p : scalar
        The p-norm to apply for Minkowski, weighted and unweighted.
        Default: 2.

        w : Tensor
        The weight vector for metrics that support weights (e.g., Minkowski).

        V : Tensor
        The variance vector for standardized Euclidean.
        Default: var(vstack([XA, XB]), axis=0, ddof=1)

        VI : Tensor
        The inverse of the covariance matrix for Mahalanobis.
        Default: inv(cov(vstack([XA, XB].T))).T

        out : Tensor
        The output tensor
        If not None, the distance matrix Y is stored in this tensor.
        Note: metric independent, it will become a regular keyword arg in a
        future scipy version

    Returns
    -------
    Y : Tensor
        A :math:`m_A` by :math:`m_B` distance matrix is returned.
        For each :math:`i` and :math:`j`, the metric
        ``dist(u=XA[i], v=XB[j])`` is computed and stored in the
        :math:`ij` th entry.

    Raises
    ------
    ValueError
        An exception is thrown if `XA` and `XB` do not have
        the same number of columns.

    Notes
    -----
    The following are common calling conventions:

    1. ``Y = cdist(XA, XB, 'euclidean')``

       Computes the distance between :math:`m` points using
       Euclidean distance (2-norm) as the distance metric between the
       points. The points are arranged as :math:`m`
       :math:`n`-dimensional row vectors in the matrix X.

    2. ``Y = cdist(XA, XB, 'minkowski', p=2.)``

       Computes the distances using the Minkowski distance
       :math:`||u-v||_p` (:math:`p`-norm) where :math:`p \\geq 1`.

    3. ``Y = cdist(XA, XB, 'cityblock')``

       Computes the city block or Manhattan distance between the
       points.

    4. ``Y = cdist(XA, XB, 'seuclidean', V=None)``

       Computes the standardized Euclidean distance. The standardized
       Euclidean distance between two n-vectors ``u`` and ``v`` is

       .. math::

          \\sqrt{\\sum {(u_i-v_i)^2 / V[x_i]}}.

       V is the variance vector; V[i] is the variance computed over all
       the i'th components of the points. If not passed, it is
       automatically computed.

    5. ``Y = cdist(XA, XB, 'sqeuclidean')``

       Computes the squared Euclidean distance :math:`||u-v||_2^2` between
       the vectors.

    6. ``Y = cdist(XA, XB, 'cosine')``

       Computes the cosine distance between vectors u and v,

       .. math::

          1 - \\frac{u \\cdot v}
                   {{||u||}_2 {||v||}_2}

       where :math:`||*||_2` is the 2-norm of its argument ``*``, and
       :math:`u \\cdot v` is the dot product of :math:`u` and :math:`v`.

    7. ``Y = cdist(XA, XB, 'correlation')``

       Computes the correlation distance between vectors u and v. This is

       .. math::

          1 - \\frac{(u - \\bar{u}) \\cdot (v - \\bar{v})}
                   {{||(u - \\bar{u})||}_2 {||(v - \\bar{v})||}_2}

       where :math:`\\bar{v}` is the mean of the elements of vector v,
       and :math:`x \\cdot y` is the dot product of :math:`x` and :math:`y`.


    8. ``Y = cdist(XA, XB, 'hamming')``

       Computes the normalized Hamming distance, or the proportion of
       those vector elements between two n-vectors ``u`` and ``v``
       which disagree. To save memory, the matrix ``X`` can be of type
       boolean.

    9. ``Y = cdist(XA, XB, 'jaccard')``

       Computes the Jaccard distance between the points. Given two
       vectors, ``u`` and ``v``, the Jaccard distance is the
       proportion of those elements ``u[i]`` and ``v[i]`` that
       disagree where at least one of them is non-zero.

    10. ``Y = cdist(XA, XB, 'chebyshev')``

       Computes the Chebyshev distance between the points. The
       Chebyshev distance between two n-vectors ``u`` and ``v`` is the
       maximum norm-1 distance between their respective elements. More
       precisely, the distance is given by

       .. math::

          d(u,v) = \\max_i {|u_i-v_i|}.

    11. ``Y = cdist(XA, XB, 'canberra')``

       Computes the Canberra distance between the points. The
       Canberra distance between two points ``u`` and ``v`` is

       .. math::

         d(u,v) = \\sum_i \\frac{|u_i-v_i|}
                              {|u_i|+|v_i|}.

    12. ``Y = cdist(XA, XB, 'braycurtis')``

       Computes the Bray-Curtis distance between the points. The
       Bray-Curtis distance between two points ``u`` and ``v`` is


       .. math::

            d(u,v) = \\frac{\\sum_i (|u_i-v_i|)}
                          {\\sum_i (|u_i+v_i|)}

    13. ``Y = cdist(XA, XB, 'mahalanobis', VI=None)``

       Computes the Mahalanobis distance between the points. The
       Mahalanobis distance between two points ``u`` and ``v`` is
       :math:`\\sqrt{(u-v)(1/V)(u-v)^T}` where :math:`(1/V)` (the ``VI``
       variable) is the inverse covariance. If ``VI`` is not None,
       ``VI`` will be used as the inverse covariance matrix.

    14. ``Y = cdist(XA, XB, 'yule')``

       Computes the Yule distance between the boolean
       vectors. (see `yule` function documentation)

    15. ``Y = cdist(XA, XB, 'matching')``

       Synonym for 'hamming'.

    16. ``Y = cdist(XA, XB, 'dice')``

       Computes the Dice distance between the boolean vectors. (see
       `dice` function documentation)

    17. ``Y = cdist(XA, XB, 'kulsinski')``

       Computes the Kulsinski distance between the boolean
       vectors. (see `kulsinski` function documentation)

    18. ``Y = cdist(XA, XB, 'rogerstanimoto')``

       Computes the Rogers-Tanimoto distance between the boolean
       vectors. (see `rogerstanimoto` function documentation)

    19. ``Y = cdist(XA, XB, 'russellrao')``

       Computes the Russell-Rao distance between the boolean
       vectors. (see `russellrao` function documentation)

    20. ``Y = cdist(XA, XB, 'sokalmichener')``

       Computes the Sokal-Michener distance between the boolean
       vectors. (see `sokalmichener` function documentation)

    21. ``Y = cdist(XA, XB, 'sokalsneath')``

       Computes the Sokal-Sneath distance between the vectors. (see
       `sokalsneath` function documentation)


    22. ``Y = cdist(XA, XB, 'wminkowski', p=2., w=w)``

       Computes the weighted Minkowski distance between the
       vectors. (see `wminkowski` function documentation)

    23. ``Y = cdist(XA, XB, f)``

       Computes the distance between all pairs of vectors in X
       using the user supplied 2-arity function f. For example,
       Euclidean distance between the vectors could be computed
       as follows::

         dm = cdist(XA, XB, lambda u, v: np.sqrt(((u-v)**2).sum()))

       Note that you should avoid passing a reference to one of
       the distance functions defined in this library. For example,::

         dm = cdist(XA, XB, sokalsneath)

       would calculate the pair-wise distances between the vectors in
       X using the Python function `sokalsneath`. This would result in
       sokalsneath being called :math:`{n \\choose 2}` times, which
       is inefficient. Instead, the optimized C version is more
       efficient, and we call it using the following syntax::

         dm = cdist(XA, XB, 'sokalsneath')

    Examples
    --------
    Find the Euclidean distances between four 2-D coordinates:

    >>> from mars.tensor.spatial import distance
    >>> coords = [(35.0456, -85.2672),
    ...           (35.1174, -89.9711),
    ...           (35.9728, -83.9422),
    ...           (36.1667, -86.7833)]
    >>> distance.cdist(coords, coords, 'euclidean').execute()
    array([[ 0.    ,  4.7044,  1.6172,  1.8856],
           [ 4.7044,  0.    ,  6.0893,  3.3561],
           [ 1.6172,  6.0893,  0.    ,  2.8477],
           [ 1.8856,  3.3561,  2.8477,  0.    ]])


    Find the Manhattan distance from a 3-D point to the corners of the unit
    cube:

    >>> import mars.tensor as mt
    >>> a = mt.array([[0, 0, 0],
    ...               [0, 0, 1],
    ...               [0, 1, 0],
    ...               [0, 1, 1],
    ...               [1, 0, 0],
    ...               [1, 0, 1],
    ...               [1, 1, 0],
    ...               [1, 1, 1]])
    >>> b = mt.array([[ 0.1,  0.2,  0.4]])
    >>> distance.cdist(a, b, 'cityblock').execute()
    array([[ 0.7],
           [ 0.9],
           [ 1.3],
           [ 1.5],
           [ 1.5],
           [ 1.7],
           [ 2.1],
           [ 2.3]])

    """
    XA = astensor(XA, order='C')
    XB = astensor(XB, order='C')

    if XA.issparse() or XB.issparse():
        raise ValueError('Sparse tensors are not supported by this function.')

    s = XA.shape
    sB = XB.shape

    if len(s) != 2:
        raise ValueError('XA must be a 2-dimensional array.')
    if len(sB) != 2:
        raise ValueError('XB must be a 2-dimensional array.')
    if s[1] != sB[1]:
        raise ValueError('XA and XB must have the same number of columns '
                         '(i.e. feature dimension.)')

    mA = s[0]
    mB = sB[0]
    out = kwargs.pop("out", None)
    if out is not None:
        if not hasattr(out, 'shape'):
            raise TypeError('return arrays must be a tensor')
        if out.shape != (mA, mB):
            raise ValueError("Output tensor has incorrect shape.")
        if out.dtype != np.double:
            raise ValueError("Output tensor must be double type.")

    if not isinstance(metric, str) and not callable(metric):
        raise TypeError('3rd argument metric must be a string identifier '
                        'or a function.')

    p = kwargs.pop('p', None)
    w = kwargs.pop('w', None)
    if w is not None:
        w = astensor(w)
    v = kwargs.pop('V', None)
    if v is not None:
        v = astensor(v)
    vi = kwargs.pop('VI', None)
    if vi is not None:
        vi = astensor(vi)

    if len(kwargs) > 0:
        raise TypeError(f"`cdist` got an unexpected keyword argument '{next(iter(kwargs))}'")

    op = TensorCdist(metric=metric,
                     p=p, w=w, v=v, vi=vi, dtype=np.dtype(float))
    shape = (XA.shape[0], XB.shape[0])
    ret = op(XA, XB, shape)

    if out is None:
        return ret
    else:
        out.data = ret.data
        return out



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

from ... import opcodes as OperandDef
from ...serialize import SeriesField, DataTypeField
from ...config import options
from ...core import OutputType
from ...tensor.utils import get_chunk_slices
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index, decide_series_chunk_size


class SeriesDataSource(DataFrameOperand, DataFrameOperandMixin):
    """
    Represents data from pandas Series
    """

    _op_type_ = OperandDef.SERIES_DATA_SOURCE

    _data = SeriesField('data')
    _dtype = DataTypeField('dtype')

    def __init__(self, data=None, dtype=None, gpu=None, sparse=None, **kw):
        if dtype is None and data is not None:
            dtype = data.dtype
        super().__init__(_data=data, _dtype=dtype, _gpu=gpu, _sparse=sparse,
                         _output_types=[OutputType.series], **kw)

    @property
    def data(self):
        return self._data

    @property
    def dtype(self):
        return self._dtype

    def __call__(self, shape, chunk_size=None):
        return self.new_series(None, shape=shape, dtype=self._dtype,
                               index_value=parse_index(self._data.index),
                               name=self._data.name, raw_chunk_size=chunk_size)

    @classmethod
    def tile(cls, op):
        series = op.outputs[0]
        raw_series = op.data

        memory_usage = raw_series.memory_usage(index=False, deep=True)
        chunk_size = series.extra_params.raw_chunk_size or options.chunk_size
        chunk_size = decide_series_chunk_size(series.shape, chunk_size, memory_usage)
        chunk_size_idxes = (range(len(size)) for size in chunk_size)

        out_chunks = []
        for chunk_shape, chunk_idx in zip(itertools.product(*chunk_size),
                                          itertools.product(*chunk_size_idxes)):
            chunk_op = op.copy().reset_key()
            slc = get_chunk_slices(chunk_size, chunk_idx)
            chunk_op._data = raw_series.iloc[slc]
            chunk_op._dtype = chunk_op._data.dtype
            out_chunk = chunk_op.new_chunk(None, shape=chunk_shape, dtype=op.dtype, index=chunk_idx,
                                           index_value=parse_index(chunk_op.data.index),
                                           name=series.name)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_seriess(None, series.shape, dtype=op.dtype,
                                  index_value=series.index_value,
                                  name=series.name, chunks=out_chunks, nsplits=chunk_size)

    @classmethod
    def execute(cls, ctx, op):
        ctx[op.outputs[0].key] = op.data


def from_pandas(data, chunk_size=None, gpu=False, sparse=False):
    op = SeriesDataSource(data=data, gpu=gpu, sparse=sparse)
    return op(data.shape, chunk_size=chunk_size)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import OutputType
from ...custom_log import redirect_custom_log
from ...serialize import BoolField, TupleField, DictField, AnyField, StringField
from ...utils import enter_current_session, quiet_stdio
from ..operands import DataFrameOperandMixin, DataFrameOperand
from ..utils import build_empty_df, build_empty_series, parse_index


class GroupByTransform(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.TRANSFORM
    _op_module_ = 'dataframe.groupby'

    _func = AnyField('func')
    _args = TupleField('args')
    _kwds = DictField('kwds')

    _call_agg = BoolField('call_agg')

    # for chunk
    _tileable_op_key = StringField('tileable_op_key')

    def __init__(self, func=None, args=None, kwds=None, call_agg=None, output_types=None,
                 tileable_op_key=None, **kw):
        super().__init__(_func=func, _args=args, _kwds=kwds, _call_agg=call_agg,
                         _output_types=output_types, _tileable_op_key=tileable_op_key, **kw)

    @property
    def func(self):
        return self._func

    @property
    def args(self):
        return getattr(self, '_args', None) or ()

    @property
    def kwds(self):
        return getattr(self, '_kwds', None) or dict()

    @property
    def call_agg(self):
        return self._call_agg

    @property
    def tileable_op_key(self):
        return self._tileable_op_key

    def _infer_df_func_returns(self, in_groupby, dtypes, index):
        index_value, output_types, new_dtypes = None, None, None

        output_types = [OutputType.dataframe] \
            if in_groupby.op.output_types[0] == OutputType.dataframe_groupby else [OutputType.series]

        try:
            mock_groupby = in_groupby.op.build_mock_groupby()
            with np.errstate(all='ignore'), quiet_stdio():
                if self.call_agg:
                    infer_df = mock_groupby.agg(self.func, *self.args, **self.kwds)
                else:
                    infer_df = mock_groupby.transform(self.func, *self.args, **self.kwds)

            # todo return proper index when sort=True is implemented
            index_value = parse_index(None, in_groupby.key, self.func)

            if isinstance(infer_df, pd.DataFrame):
                output_types = [OutputType.dataframe]
                new_dtypes = new_dtypes or infer_df.dtypes
            else:
                output_types = [OutputType.series]
                new_dtypes = new_dtypes or (infer_df.name, infer_df.dtype)
        except:  # noqa: E722  # nosec
            pass

        self.output_types = output_types if not self.output_types else self.output_types
        dtypes = new_dtypes if dtypes is None else dtypes
        index_value = index_value if index is None else parse_index(index)
        return dtypes, index_value

    def __call__(self, groupby, dtypes=None, dtype=None, name=None, index=None):
        in_df = groupby.inputs[0]

        if dtypes is None and dtype is not None:
            dtypes = (name, dtype)
        dtypes, index_value = self._infer_df_func_returns(groupby, dtypes, index)
        if index_value is None:
            index_value = parse_index(None, (in_df.key, in_df.index_value.key))
        for arg, desc in zip((self.output_types, dtypes), ('output_types', 'dtypes')):
            if arg is None:
                raise TypeError(f'Cannot determine {desc} by calculating with enumerate data, '
                                'please specify it as arguments')

        if self.output_types[0] == OutputType.dataframe:
            new_shape = (np.nan if self.call_agg else in_df.shape[0], len(dtypes))
            return self.new_dataframe([groupby], shape=new_shape, dtypes=dtypes, index_value=index_value,
                                      columns_value=parse_index(dtypes.index, store_data=True))
        else:
            name, dtype = dtypes
            new_shape = (np.nan,) if self.call_agg else groupby.shape
            return self.new_series([groupby], name=name, shape=new_shape, dtype=dtype,
                                   index_value=index_value)

    @classmethod
    def tile(cls, op):
        in_groupby = op.inputs[0]
        out_df = op.outputs[0]

        chunks = []
        for c in in_groupby.chunks:
            inp_chunks = [c]

            new_op = op.copy().reset_key()
            new_op._tileable_op_key = op.key
            if op.output_types[0] == OutputType.dataframe:
                new_index = c.index if c.ndim == 2 else c.index + (0,)
                chunks.append(new_op.new_chunk(
                    inp_chunks, index=new_index, shape=(np.nan, len(out_df.dtypes)), dtypes=out_df.dtypes,
                    columns_value=out_df.columns_value, index_value=out_df.index_value))
            else:
                chunks.append(new_op.new_chunk(
                    inp_chunks, name=out_df.name, index=(c.index[0],), shape=(np.nan,), dtype=out_df.dtype,
                    index_value=out_df.index_value))

        new_op = op.copy()
        kw = out_df.params.copy()
        kw['chunks'] = chunks
        if op.output_types[0] == OutputType.dataframe:
            kw['nsplits'] = ((np.nan,) * len(chunks), (len(out_df.dtypes),))
        else:
            kw['nsplits'] = ((np.nan,) * len(chunks),)
        return new_op.new_tileables([in_groupby], **kw)

    @classmethod
    @redirect_custom_log
    @enter_current_session
    def execute(cls, ctx, op):
        in_data = ctx[op.inputs[0].key]
        out_chunk = op.outputs[0]

        if not in_data:
            if op.output_types[0] == OutputType.dataframe:
                ctx[op.outputs[0].key] = build_empty_df(out_chunk.dtypes)
            else:
                ctx[op.outputs[0].key] = build_empty_series(
                    out_chunk.dtype, name=out_chunk.name)
            return

        if op.call_agg:
            result = in_data.agg(op.func, *op.args, **op.kwds)
        elif in_data.shape[0] > 0:
            # cannot perform groupby-transform over empty dataframe
            result = in_data.transform(op.func, *op.args, **op.kwds)
        else:
            if out_chunk.ndim == 2:
                result = pd.DataFrame(columns=out_chunk.dtypes.index)
            else:
                result = pd.Series([], name=out_chunk.name, dtype=out_chunk.dtype)

        if result.ndim == 2:
            result = result.astype(out_chunk.dtypes, copy=False)
        else:
            result = result.astype(out_chunk.dtype, copy=False)
        ctx[op.outputs[0].key] = result


def groupby_transform(groupby, f, *args, dtypes=None, dtype=None, name=None, index=None,
                      output_types=None, **kwargs):
    """
    Call function producing a like-indexed DataFrame on each group and
    return a DataFrame having the same indexes as the original object
    filled with the transformed values

    Parameters
    ----------
    f : function
        Function to apply to each group.

    dtypes : Series, default None
        Specify dtypes of returned DataFrames. See `Notes` for more details.

    dtype : numpy.dtype, default None
        Specify dtype of returned Series. See `Notes` for more details.

    name : str, default None
        Specify name of returned Series. See `Notes` for more details.

    *args
        Positional arguments to pass to func

    **kwargs
        Keyword arguments to be passed into func.

    Returns
    -------
    DataFrame

    See Also
    --------
    DataFrame.groupby.apply
    DataFrame.groupby.aggregate
    DataFrame.transform

    Notes
    -----
    Each group is endowed the attribute 'name' in case you need to know
    which group you are working on.

    The current implementation imposes three requirements on f:

    * f must return a value that either has the same shape as the input
      subframe or can be broadcast to the shape of the input subframe.
      For example, if `f` returns a scalar it will be broadcast to have the
      same shape as the input subframe.
    * if this is a DataFrame, f must support application column-by-column
      in the subframe. If f also supports application to the entire subframe,
      then a fast path is used starting from the second chunk.
    * f must not mutate groups. Mutation is not supported and may
      produce unexpected results.

    Notes
    -----
    When deciding output dtypes and shape of the return value, Mars will
    try applying ``func`` onto a mock grouped object, and the transform call
    may fail.

    * For DataFrame output, you need to specify a list or a pandas Series
      as ``dtypes`` of output DataFrame. ``index`` of output can also be
      specified.
    * For Series output, you need to specify ``dtype`` and ``name`` of
      output Series.

    Examples
    --------

    >>> import mars.dataframe as md
    >>> df = md.DataFrame({'A' : ['foo', 'bar', 'foo', 'bar',
    ...                           'foo', 'bar'],
    ...                    'B' : ['one', 'one', 'two', 'three',
    ...                           'two', 'two'],
    ...                    'C' : [1, 5, 5, 2, 5, 5],
    ...                    'D' : [2.0, 5., 8., 1., 2., 9.]})
    >>> grouped = df.groupby('A')
    >>> grouped.transform(lambda x: (x - x.mean()) / x.std()).execute()
              C         D
    0 -1.154701 -0.577350
    1  0.577350  0.000000
    2  0.577350  1.154701
    3 -1.154701 -1.000000
    4  0.577350 -0.577350
    5  0.577350  1.000000

    Broadcast result of the transformation

    >>> grouped.transform(lambda x: x.max() - x.min()).execute()
       C    D
    0  4  6.0
    1  3  8.0
    2  4  6.0
    3  3  8.0
    4  4  6.0
    5  3  8.0
    """
    call_agg = kwargs.pop('_call_agg', False)
    if not call_agg and isinstance(f, (dict, list)):
        raise TypeError(f'Does not support transform with {type(f)}')

    op = GroupByTransform(func=f, args=args, kwds=kwargs, output_types=output_types,
                          call_agg=call_agg)
    return op(groupby, dtypes=dtypes, dtype=dtype, name=name, index=index)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd

from ...config import options
from ...operands import OperandStage
from ...serialize import Int32Field, Int64Field, StringField, \
    ListField, BoolField, ValueType
from ...tiles import TilesError
from ...utils import ceildiv
from ..operands import DataFrameOperand
from ..utils import parse_index


class DataFrameSortOperand(DataFrameOperand):
    _axis = Int32Field('axis')
    _ascending = BoolField('ascending')
    _inplace = BoolField('inplace')
    _kind = StringField('kind')
    _na_position = StringField('na_position')
    _ignore_index = BoolField('ignore_index')
    _parallel_kind = StringField('parallel_kind')
    _psrs_kinds = ListField('psrs_kinds', ValueType.string)
    _nrows = Int64Field('nrows')

    def __init__(self, axis=None, ascending=None, inplace=None, kind=None, na_position=None,
                 ignore_index=None, parallel_kind=None, psrs_kinds=None,
                 nrows=None, gpu=False, **kw):
        super().__init__(_axis=axis, _ascending=ascending, _inplace=inplace, _kind=kind,
                         _na_position=na_position, _ignore_index=ignore_index,
                         _parallel_kind=parallel_kind, _psrs_kinds=psrs_kinds,
                         _nrows=nrows, _gpu=gpu, **kw)

    @property
    def axis(self):
        return self._axis

    @property
    def ascending(self):
        return self._ascending

    @property
    def inplace(self):
        return self._inplace

    @property
    def kind(self):
        return self._kind

    @property
    def na_position(self):
        return self._na_position

    @property
    def ignore_index(self):
        return self._ignore_index

    @property
    def parallel_kind(self):
        return self._parallel_kind

    @property
    def psrs_kinds(self):
        return self._psrs_kinds

    @property
    def nrows(self):
        return self._nrows

    @classmethod
    def _tile_head(cls, op: "DataFrameSortOperand"):
        from ..merge import DataFrameConcat

        inp = op.inputs[0]
        out = op.outputs[0]
        axis = op.axis
        assert axis == 0
        pd_index = out.index_value.to_pandas()
        combine_size = options.combine_size

        if inp.ndim == 2:
            if inp.chunk_shape[1 - axis] > 1:  # pragma: no cover
                if any(pd.isna(s) for s in inp.nsplits[1 - axis]):
                    raise TilesError('Need to be executed first '
                                     'due to unknown chunk shape')
                inp = inp.rechunk({1 - axis: inp.shape[1 - axis]})._inplace_tile()

        out_chunks = []
        for c in inp.chunks:
            chunk_op = op.copy().reset_key()
            chunk_op._stage = OperandStage.map
            chunk_params = c.params
            chunk_params['index_value'] = parse_index(pd_index, c)
            out_chunks.append(chunk_op.new_chunk([c], kws=[chunk_params]))

        while True:
            chunk_size = ceildiv(len(out_chunks), combine_size)
            combine_chunks = []
            for i in range(chunk_size):
                chunk_index = (i,) if inp.ndim == 1 else (i, 0)

                to_combine_chunks = out_chunks[i * combine_size: (i + 1) * combine_size]
                concat_params = to_combine_chunks[0].params
                concat_params['index'] = chunk_index
                shape = list(to_combine_chunks[0].shape)
                shape[0] = sum(c.shape[0] for c in to_combine_chunks)
                shape = tuple(shape)
                concat_params['shape'] = shape
                c = DataFrameConcat(axis=axis, output_types=op.output_types).new_chunk(
                    to_combine_chunks, kws=[concat_params])

                chunk_op = op.copy().reset_key()
                chunk_op._stage = OperandStage.combine \
                    if chunk_size > 1 else OperandStage.agg
                chunk_params = c.params
                chunk_params['index_value'] = parse_index(pd_index, c)
                chunk_params['shape'] = (min(shape[0], op.nrows),) + shape[1:]
                combine_chunks.append(chunk_op.new_chunk([c], kws=[chunk_params]))
            out_chunks = combine_chunks
            if chunk_size == 1:
                break

        new_op = op.copy()
        params = out.params
        params['nsplits'] = tuple((s,) for s in out.shape)
        params['chunks'] = out_chunks
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def _tile(cls, op):  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def tile(cls, op: "DataFrameSortOperand"):
        if op.nrows is not None:
            return cls._tile_head(op)
        else:
            return cls._tile(op)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

NULL = 0

# creation
# tensor
SCALAR = 1
TENSOR_DATA_SOURCE = 2
TENSOR_ONES = 3
TENSOR_ONES_LIKE = 4
TENSOR_ZEROS = 5
TENSOR_ZEROS_LIKE = 6
TENSOR_EMPTY = 7
TENSOR_EMPTY_LIKE = 8
TENSOR_FULL = 9
TENSOR_FULL_LIKE = 25
TENSOR_ARANGE = 10
TENSOR_INDICES = 11
TENSOR_DIAG = 12
TENSOR_EYE = 13
TENSOR_LINSPACE = 14
TENSOR_TRIU = 15
TENSOR_TRIL = 16
# external storage
TENSOR_FROM_TILEDB = 18
TENSOR_STORE_TILEDB = 19
TENSOR_STORE_TILEDB_CONSOLIDATE = 20
TENSOR_FROM_DATAFRAME = 22
TENSOR_FROM_HDF5 = 27
TENSOR_STORE_HDF5 = 28
TENSOR_FROM_ZARR = 29
TENSOR_STORE_ZARR = 32

# dataframe
DATAFRAME_DATA_SOURCE = 17
DATAFRAME_FROM_TENSOR = 21
DATAFRAME_FROM_RECORDS = 24
# series
SERIES_DATA_SOURCE = 23
SERIES_FROM_TENSOR = 26
SERIES_FROM_INDEX = 39
# index
INDEX_DATA_SOURCE = 33
DATE_RANGE = 34
TIMEDELTA_RANGE = 35
CHECK_MONOTONIC = 38
# misc
MEMORY_USAGE = 36
REBALANCE = 37

# GPU
TO_GPU = 30
TO_CPU = 31

# random
RAND_RAND = 41
RAND_RANDN = 42
RAND_RANDINT = 43
RAND_RANDOM_INTEGERS = 44
RAND_RANDOM_SAMPLE = 45
RAND_RANDOM = 46
RAND_RANF = 47
RAND_SAMPLE = 48
RAND_BYTES = 49

# random distribution
RAND_BETA = 50
RAND_BINOMIAL = 51
RAND_CHISQUARE = 52
RAND_CHOICE = 53
RAND_DIRICHLET = 54
RAND_EXPONENTIAL = 55
RAND_F = 56
RAND_GAMMA = 57
RAND_GEOMETRIC = 58
RAND_GUMBEL = 59
RAND_HYPERGEOMETRIC = 60
RAND_LAPLACE = 61
RAND_LOGISTIC = 62
RAND_LOGNORMAL = 63
RAND_LOGSERIES = 64
RAND_MULTINOMIAL = 65
RAND_MULTIVARIATE_NORMAL = 66
RAND_NEGATIVE_BINOMIAL = 67
RAND_NONCENTRAL_CHISQURE = 68
RAND_NONCENTRAL_F = 69
RAND_NORMAL = 70
RAND_PARETO = 71
RAND_PERMUTATION = 72
RAND_POSSION = 73
RAND_POWER = 74
RAND_RAYLEIGH = 75
RAND_SHUFFLE = 76
RAND_STANDARD_CAUCHY = 77
RAND_STANDARD_EXPONENTIAL = 78
RAND_STANDARD_GAMMMA = 79
RAND_STANDARD_NORMAL = 80
RAND_STANDARD_T = 81
RAND_TOMAXINT = 82
RAND_TRIANGULAR = 83
RAND_UNIFORM = 84
RAND_VONMISES = 85
RAND_WALD = 86
RAND_WEIBULL = 87
RAND_ZIPF = 88
PERMUTATION = 89
UNIQUE = 90

# ufunc
ADD = 101
SUB = 102
MUL = 103
DIV = 104
TRUEDIV = 105
FLOORDIV = 106
POW = 107
MOD = 108
FMOD = 109
LOGADDEXP = 110
LOGADDEXP2 = 111
NEGATIVE = 112
POSITIVE = 113
ABSOLUTE = 114
FABS = 115
ABS = 116
RINT = 117
SIGN = 118
CONJ = 119
EXP = 120
EXP2 = 121
LOG = 122
LOG2 = 123
LOG10 = 124
EXPM1 = 125
LOG1P = 126
SQRT = 127
SQUARE = 128
CBRT = 129
RECIPROCAL = 130
EQ = 131
NE = 132
LT = 133
LE = 134
GT = 135
GE = 136
SIN = 137
COS = 138
TAN = 139
ARCSIN = 140
ARCCOS = 141
ARCTAN = 142
ARCTAN2 = 143
HYPOT = 144
SINH = 145
COSH = 146
TANH = 147
ARCSINH = 148
ARCCOSH = 149
ARCTANH = 150
DEG2RAD = 151
RAD2DEG = 152
BITAND = 153
BITOR = 154
BITXOR = 155
INVERT = 156
LSHIFT = 157
RSHIFT = 158
AND = 159
OR = 160
XOR = 161
NOT = 162
MAXIMUM = 163
MINIMUM = 164
AROUND = 165
FLOAT_POWER = 166
FMAX = 167
FMIN = 168
ISFINITE = 169
ISINF = 170
ISNAN = 171
SIGNBIT = 172
COPYSIGN = 173
NEXTAFTER = 174
SPACING = 175
LDEXP = 176
FREXP = 177
MODF = 178
FLOOR = 179
CEIL = 180
TRUNC = 181
DEGREES = 182
RADIANS = 183
CLIP = 184
ISREAL = 185
ISCOMPLEX = 186
REAL = 187
IMAG = 188
FIX = 189
I0 = 190
SINC = 191
NAN_TO_NUM = 192
ISCLOSE = 193
DIVMOD = 194
ANGLE = 195
SET_REAL = 196
SET_IMAG = 197

# special
SPECIAL = 200
ERF = 201
ENTR = 202
REL_ENTR = 203

# spatial
PDIST = 231
CDIST = 232
SQUAREFORM = 233

# tree operand
TREE_ADD = 251
TREE_MULTIPLY = 252

# reduction
CUMSUM = 301
CUMPROD = 302
PROD = 303
SUM = 304
MAX = 305
MIN = 306
ALL = 307
ANY = 308
MEAN = 309
ARGMAX = 310
ARGMIN = 311
NANSUM = 312
NANMAX = 313
NANMIN = 314
NANPROD = 315
NANMEAN = 316
NANARGMAX = 317
NANARGMIN = 318
COUNT_NONZERO = 319
MOMENT = 320
NANMOMENT = 321
VAR = 322
STD = 323
NANVAR = 324
NANSTD = 325
NANCUMSUM = 326
NANCUMPROD = 327
COUNT = 343
CUMMAX = 344
CUMMIN = 345
CUMCOUNT = 346
CORR = 347
REDUCTION_SIZE = 348
CUSTOM_REDUCTION = 349
SKEW = 350
KURTOSIS = 351
SEM = 352
STR_CONCAT = 353
MAD = 354

# tensor operand
RESHAPE = 401
SLICE = 402
INDEX = 403
INDEXSETVALUE = 404
CONCATENATE = 405
RECHUNK = 406
ASTYPE = 407
TRANSPOSE = 408
SWAPAXES = 409
BROADCAST_TO = 410
STACK = 411
WHERE = 412
CHOOSE = 413
NONZERO = 414
ARGWHERE = 415
UNRAVEL_INDEX = 416
RAVEL_MULTI_INDEX = 417
ARRAY_SPLIT = 418
SQUEEZE = 419
DIGITIZE = 420
REPEAT = 421
COPYTO = 422
ISIN = 423
SEARCHSORTED = 428
SORT = 429
HISTOGRAM = 430
HISTOGRAM_BIN_EDGES = 431
PARTITION = 432
QUANTILE = 440
FILL_DIAGONAL = 441
NORMALIZE = 442
TOPK = 443
TRAPZ = 444
GET_SHAPE = 445
# fancy index, distributed phase is a shuffle operation that
# the fancy indexes will be distributed to the left chunks
# the concat phase will concat back the indexed left chunks and index
# according to the original fancy index order
FANCY_INDEX_DISTRIBUTE = 424
FANCY_INDEX_CONCAT = 425

# linear algebra
TENSORDOT = 501
DOT = 502
MATMUL = 503
CHOLESKY = 510
QR = 511
SVD = 512
LU = 513
SOLVE_TRIANGULAR = 520
INV = 521
NORM = 530

# fft
FFT = 601
IFFT = 602
FFT2 = 603
IFFT2 = 604
FFTN = 605
IFFTN = 606
RFFT = 607
IRFFT = 608
RFFT2 = 609
IRFFT2 = 610
RFFTN = 611
IRFFTN = 612
HFFT = 613
IHFFT = 614
FFTFREQ = 615
FFTFREQ_CHUNK = 616
RFFTFREQ = 617
FFTSHIFT = 618
IFFTSHIFT = 619

# einsum
EINSUM = 630

# sparse creation
SPARSE_MATRIX_DATA_SOURCE = 701
DENSE_TO_SPARSE = 702
SPARSE_TO_DENSE = 703

# DataFrame
MAP = 710
DESCRIBE = 712
FILL_NA = 713
AGGREGATE = 714
STRING_METHOD = 715
DATETIME_METHOD = 716
APPLY = 717
TRANSFORM = 718
CHECK_NA = 719
DROP_NA = 720
NUNIQUE = 721
CUT = 722
SHIFT = 723
DATAFRAME_DIFF = 724
VALUE_COUNTS = 725
TO_DATETIME = 726
DATAFRAME_DROP = 727
DROP_DUPLICATES = 728
MELT = 729
RENAME = 731
INSERT = 732
MAP_CHUNK = 733
CARTESIAN_CHUNK = 734
EXPLODE = 735
REPLACE = 736
RENAME_AXIS = 737
DATAFRAME_EVAL = 738
DUPLICATED = 739
DELETE = 740

FUSE = 801

# control
ENTER = 901
LEAVE = 902
FIX_LATEST = 903
IF_ELSE = 904
NEXT_ITER = 905

# table like input for tensor
TABLE_COO = 1003
# store tensor as coo format
STORE_COO = 1004

# shuffle
SHUFFLE_PROXY = 2001
DATAFRAME_INDEX_ALIGN = 2004
UNIQUE_INVERSE_REDUCE = 2013

# indexing
DATAFRAME_SET_INDEX = 2020
DATAFRAME_SET_AXIS = 730
DATAFRAME_ILOC_GETITEM = 2021
DATAFRAME_ILOC_SETITEM = 2022
DATAFRAME_LOC_GETITEM = 2023
DATAFRAME_LOC_SETITEM = 2024

# merge
DATAFRAME_SHUFFLE_MERGE = 2010
DATAFRAME_SHUFFLE_MERGE_ALIGN = 2011

# append
APPEND = 2015

# reset index
RESET_INDEX = 2028
# reindex
REINDEX = 2029

# groupby
GROUPBY = 2030
GROUPBY_AGG = 2033
GROUPBY_CONCAT = 2034
GROUPBY_HEAD = 2035
GROUPBY_SAMPLE_ILOC = 2036

# parallel sorting by regular sampling
PSRS_SORT_REGULAR_SMAPLE = 2040
PSRS_CONCAT_PIVOT = 2041
PSRS_SHUFFLE = 2042
PSRS_ALIGN = 2043
# partition
CALC_PARTITIONS_INFO = 2046
PARTITION_MERGED = 2047

# dataframe sort
SORT_VALUES = 2050
SORT_INDEX = 2051

# window
ROLLING_AGG = 2060
EXPANDING_AGG = 2061
EWM_AGG = 2062

# store
READ_CSV = 2100
TO_CSV = 2101
READ_PARQUET = 2103
TO_PARQUET = 2104
READ_SQL = 2105
TO_SQL = 2108

TO_CSV_STAT = 2102

# standardize range index
STANDARDIZE_RANGE_INDEX = 2107

# successors exclusive
SUCCESSORS_EXCLUSIVE = 2002

# read images
IMREAD = 2110

# machine learning

# pairwise distances
PAIRWISE_EUCLIDEAN_DISTANCES = 2200
PAIRWISE_MANHATTAN_DISTANCES = 2201
PAIRWISE_COSINE_DISTANCES = 2202
PAIRWISE_HAVERSINE_DISTANCES = 2203
PAIRWISE_DISTANCES_TOPK = 2204

# nearest neighbors
KD_TREE_TRAIN = 2230
KD_TREE_QUERY = 2231
BALL_TREE_TRAIN = 2232
BALL_TREE_QUERY = 2233
FAISS_BUILD_INDEX = 2234
FAISS_TRAIN_SAMPLED_INDEX = 2235
FAISS_QUERY = 2236
PROXIMA_SIMPLE_BUILDER = 2238
PROXIMA_SIMPLE_SEARCHER = 2239
KNEIGHBORS_GRAPH = 2237

# CLUSTER
KMEANS_PLUS_PLUS_INIT = 2250
KMEANS_SCALABLE_PLUS_PLUS_INIT = 2251
KMEANS_ELKAN_INIT_BOUNDS = 2252
KMEANS_ELKAN_UPDATE = 2253
KMEANS_ELKAN_POSTPROCESS = 2254
KMEANS_LLOYD_UPDATE = 2255
KMEANS_LLOYD_POSTPROCESS = 2256
KMEANS_INERTIA = 2257
KMEANS_RELOCASTE_EMPTY_CLUSTERS = 2258

# XGBoost
XGBOOST_TRAIN = 3001
XGBOOST_PREDICT = 3002
TO_DMATRIX = 3003
START_TRACKER = 3004

# LightGBM
LGBM_TRAIN = 3020
LGBM_PREDICT = 3021
LGBM_ALIGN = 3022

# TensorFlow
RUN_TENSORFLOW = 3010

# PyTorch
RUN_PYTORCH = 3011

# statsmodels
STATSMODELS_TRAIN = 3012
STATSMODELS_PREDICT = 3013

# learn
# checks
CHECK_NON_NEGATIVE = 3300
# classifier check targets
CHECK_TARGETS = 3301
ASSERT_ALL_FINITE = 3302
# multilabel
IS_MULTILABEL = 3303
# get type
TYPE_OF_TARGET = 3304
# classification
ACCURACY_SCORE = 3305
# port detection
COLLECT_PORTS = 3306

# Remote Functions and class
REMOTE_FUNCATION = 5001
RUN_SCRIPT = 5002

# vineyard
TENSOR_FROM_VINEYARD = 4002
TENSOR_STORE_VINEYARD_CHUNK = 4003
TENSOR_STORE_VINEYARD_GLOBAL_META = 4005
DATAFRAME_FROM_VINEYARD = 4007
DATAFRAME_STORE_VINEYARD_CHUNK = 4008
DATAFRAME_STORE_VINEYARD_GLOBAL_META = 4010

CHOLESKY_FUSE = 999988

# fetches
FETCH_SHUFFLE = 999998
FETCH = 999999


_val_to_dict = dict()
for _var_name, _var_val in globals().copy().items():
    if not isinstance(_var_val, int):
        continue
    if _var_val in _val_to_dict:  # pragma: no cover
        raise ImportError(f'Cannot import opcode: {_var_name} and '
                          f'{_val_to_dict[_var_val]} collides with value {_var_val}')
    _val_to_dict[_var_val] = _var_name
del _val_to_dict, _var_name, _var_val



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numpy.linalg import LinAlgError

from ... import opcodes as OperandDef
from ...serialize import KeyField
from ...core import ExecutableTuple
from ...utils import check_chunks_unknown_shape, recursive_tile
from ...tiles import TilesError
from ..array_utils import device, as_same_device, is_sparse_module
from ..operands import TensorHasInput, TensorOperandMixin
from ..datasource import tensor as astensor


class TensorLU(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.LU

    _input = KeyField('input')

    def __init__(self, dtype=None, sparse=False, **kw):
        super().__init__(_dtype=dtype, _sparse=sparse, **kw)

    @property
    def output_limit(self):
        return 3

    def __call__(self, a):
        import scipy.linalg

        a = astensor(a)
        if a.ndim != 2:
            raise LinAlgError(f'{a.ndim}-dimensional array given. '
                              'Tensor must be two-dimensional')

        if a.shape[0] > a.shape[1]:
            p_shape = (a.shape[0],) * 2
            l_shape = a.shape
            u_shape = (a.shape[1],) * 2
        elif a.shape[0] < a.shape[1]:
            p_shape = (a.shape[0],) * 2
            l_shape = (a.shape[0],) * 2
            u_shape = a.shape
        else:
            p_shape, l_shape, u_shape = (a.shape,) * 3

        tiny_p, tiny_l, tiny_u = scipy.linalg.lu(np.array([[1, 2], [2, 5]], dtype=a.dtype))

        order = a.order
        p, l, u = self.new_tensors([a],
                                   kws=[
                                       {'side': 'p', 'dtype': tiny_p.dtype,
                                        'shape': p_shape, 'order': order},
                                       {'side': 'l', 'dtype': tiny_l.dtype,
                                        'shape': l_shape, 'order': order},
                                       {'side': 'u', 'dtype': tiny_u.dtype,
                                        'shape': u_shape, 'order': order},
                                   ])
        return ExecutableTuple([p, l, u])

    @classmethod
    def _tile_one_chunk(cls, op):
        p, l, u = op.outputs
        chunk_op = op.copy().reset_key()
        chunk_kws = [
            {'side': 'p', 'dtype': p.dtype,
             'shape': p.shape, 'order': p.order,
             'index': (0,) * p.ndim},
            {'side': 'l', 'dtype': l.dtype,
             'shape': l.shape, 'order': l.order,
             'index': (0,) * l.ndim},
            {'side': 'u', 'dtype': u.dtype,
             'shape': u.shape, 'order': u.order,
             'index': (0,) * u.ndim}
        ]
        chunks = chunk_op.new_chunks(op.input.chunks, kws=chunk_kws)

        new_op = op.copy()
        kws = [p.params, l.params, u.params]
        for i, out in enumerate([p, l, u]):
            kws[i]['nsplits'] = tuple((s,) for s in out.shape)
            kws[i]['chunks'] = [chunks[i]]
        return new_op.new_tensors(op.inputs, kws=kws)

    @classmethod
    def tile(cls, op):
        if len(op.input.chunks) == 1:
            return cls._tile_one_chunk(op)

        from ..arithmetic.subtract import TensorSubtract
        from ..arithmetic.add import TensorTreeAdd
        from ..base.transpose import TensorTranspose
        from ..merge.vstack import vstack
        from ..merge.hstack import hstack
        from ..datasource.zeros import TensorZeros, zeros
        from .dot import TensorDot
        from .solve_triangular import TensorSolveTriangular

        P, L, U = op.outputs
        raw_in_tensor = in_tensor = op.input
        out_tensor = op.outputs[0]

        if in_tensor.shape[0] > in_tensor.shape[1]:
            zero_tensor = zeros((in_tensor.shape[0], in_tensor.shape[0] - in_tensor.shape[1]),
                                dtype=in_tensor.dtype, sparse=in_tensor.issparse(),
                                gpu=in_tensor.op.gpu,
                                chunk_size=(in_tensor.nsplits[0], max(in_tensor.nsplits[1])),
                                order=in_tensor.order.value)
            in_tensor = hstack([in_tensor, zero_tensor])
            recursive_tile(in_tensor)
        elif in_tensor.shape[0] < in_tensor.shape[1]:
            zero_tensor = zeros((in_tensor.shape[1] - in_tensor.shape[0], in_tensor.shape[1]),
                                dtype=in_tensor.dtype, sparse=in_tensor.issparse(),
                                gpu=in_tensor.op.gpu,
                                chunk_size=(max(in_tensor.nsplits[0]), in_tensor.nsplits[1]),
                                order=in_tensor.order.value)
            in_tensor = vstack([in_tensor, zero_tensor])
            recursive_tile(in_tensor)

        check_chunks_unknown_shape([in_tensor], TilesError)
        if in_tensor.nsplits[0] != in_tensor.nsplits[1]:
            # all chunks on diagonal should be square
            nsplits = in_tensor.nsplits[0]
            in_tensor = in_tensor.rechunk([nsplits, nsplits])._inplace_tile()

        p_chunks, p_invert_chunks, lower_chunks, l_permuted_chunks, upper_chunks = {}, {}, {}, {}, {}
        for i in range(in_tensor.chunk_shape[0]):
            for j in range(in_tensor.chunk_shape[1]):
                if i < j:
                    chunk_shape = (in_tensor.nsplits[0][i], in_tensor.nsplits[1][j])
                    p_chunk = TensorZeros(sparse=op.sparse, order=out_tensor.order.value).new_chunk(
                        None, shape=chunk_shape, index=(i, j), order=out_tensor.order)
                    lower_chunk = TensorZeros(sparse=op.sparse, order=out_tensor.order.value).new_chunk(
                        None, shape=chunk_shape, index=(i, j), order=out_tensor.order)
                    p_chunks[p_chunk.index] = p_chunk
                    lower_chunks[lower_chunk.index] = lower_chunk

                    target_u = in_tensor.cix[i, j]
                    p_invert = p_invert_chunks[i, i]
                    target = TensorDot(dtype=U.dtype, sparse=U.op.sparse).new_chunk(
                        [p_invert, target_u], shape=(p_invert.shape[0], target_u.shape[1]),
                        order=out_tensor.order)
                    if i > 0:
                        prev_chunks_u = []
                        for p in range(i):
                            a, b = lower_chunks[i, p], upper_chunks[p, j]
                            prev_chunk = TensorDot(dtype=U.dtype, sparse=U.op.sparse).new_chunk(
                                [a, b], shape=(a.shape[0], b.shape[1]), order=out_tensor.order)
                            prev_chunks_u.append(prev_chunk)
                        if len(prev_chunks_u) == 1:
                            s = prev_chunks_u[0]
                        else:
                            tree_add_op = TensorTreeAdd(args=prev_chunks_u, dtype=prev_chunks_u[0].dtype,
                                                        sparse=op.sparse)
                            s = tree_add_op.new_chunk(prev_chunks_u, shape=prev_chunks_u[0].shape)
                        target = TensorSubtract(dtype=U.dtype, lhs=target, rhs=s,
                                                order=out_tensor.order.value).new_chunk(
                            [target, s], shape=target.shape, order=out_tensor.order)
                    upper_chunk = TensorSolveTriangular(lower=True, dtype=U.dtype, strict=False,
                                                        sparse=lower_chunks[i, i].op.sparse).new_chunk(
                        [lower_chunks[i, i], target], shape=target.shape, index=(i, j),
                        order=out_tensor.order)
                    upper_chunks[upper_chunk.index] = upper_chunk
                elif i == j:
                    target = in_tensor.cix[i, j]
                    if i > 0:
                        prev_chunks = []
                        for p in range(i):
                            a, b = l_permuted_chunks[i, p], upper_chunks[p, j]
                            prev_chunk = TensorDot(dtype=a.dtype, sparse=op.sparse).new_chunk(
                                [a, b], shape=(a.shape[0], b.shape[1]), order=out_tensor.order)
                            prev_chunks.append(prev_chunk)
                        if len(prev_chunks) == 1:
                            s = prev_chunks[0]
                        else:
                            tree_add_op = TensorTreeAdd(args=prev_chunks, dtype=prev_chunks[0].dtype,
                                                        sparse=op.sparse)
                            s = tree_add_op.new_chunk(prev_chunks, shape=prev_chunks[0].shape)
                        target = TensorSubtract(dtype=L.dtype, lhs=target, rhs=s,
                                                order=out_tensor.order.value).new_chunk(
                            [target, s], shape=target.shape)
                    new_op = TensorLU(dtype=op.dtype, sparse=target.op.sparse)
                    lu_chunks = new_op.new_chunks([target],
                                                  index=(i, j),
                                                  order=out_tensor.order,
                                                  kws=[
                                                      {'side': 'p', 'dtype': P.dtype, 'shape': target.shape},
                                                      {'side': 'l', 'dtype': L.dtype, 'shape': target.shape},
                                                      {'side': 'u', 'dtype': U.dtype, 'shape': target.shape},
                                                  ])
                    p_chunk, lower_chunk, upper_chunk = lu_chunks
                    # transposed p equals to inverted p
                    p_chunk_invert = TensorTranspose(dtype=p_chunk.dtype, sparse=op.sparse).new_chunk(
                        [p_chunk], shape=p_chunk.shape, index=p_chunk.index, order=out_tensor.order)
                    p_chunks[p_chunk.index] = p_chunk
                    p_invert_chunks[p_chunk_invert.index] = p_chunk_invert
                    lower_chunks[lower_chunk.index] = lower_chunk
                    upper_chunks[upper_chunk.index] = upper_chunk

                    # l_permuted should be transferred to the final lower triangular
                    for p in range(i):
                        l_permuted_chunk = l_permuted_chunks[i, p]
                        l_chunk = TensorDot(dtype=L.dtype, sparse=L.op.sparse).new_chunk(
                            [p_chunk_invert, l_permuted_chunk],
                            shape=(p_chunk_invert.shape[0], l_permuted_chunk.shape[1]),
                            index=l_permuted_chunk.index, order=out_tensor.order
                        )
                        lower_chunks[l_permuted_chunk.index] = l_chunk
                else:
                    chunk_shape = (in_tensor.nsplits[0][i], in_tensor.nsplits[1][j])
                    p_chunk = TensorZeros(sparse=op.sparse, order=out_tensor.order.value).new_chunk(
                        None, shape=chunk_shape, index=(i, j), order=out_tensor.order)
                    upper_chunk = TensorZeros(sparse=op.sparse, order=out_tensor.order.value).new_chunk(
                        None, shape=chunk_shape, index=(i, j), order=out_tensor.order)
                    p_chunks[p_chunk.index] = p_chunk
                    upper_chunks[upper_chunk.index] = upper_chunk
                    target_l = in_tensor.cix[i, j]
                    if j > 0:
                        prev_chunks_l = []
                        for p in range(j):
                            a, b = l_permuted_chunks[i, p], upper_chunks[p, j]
                            prev_chunk = TensorDot(dtype=L.dtype, sparse=L.op.sparse).new_chunk(
                                [a, b], shape=(a.shape[0], b.shape[1]), order=out_tensor.order)
                            prev_chunks_l.append(prev_chunk)
                        if len(prev_chunks_l) == 1:
                            s = prev_chunks_l[0]
                        else:
                            tree_add_op = TensorTreeAdd(args=prev_chunks_l, dtype=prev_chunks_l[0].dtype,
                                                        sparse=op.sparse)
                            s = tree_add_op.new_chunk(prev_chunks_l, shape=prev_chunks_l[0].shape)
                        target_l = TensorSubtract(dtype=L.dtype, lhs=target_l, rhs=s,
                                                  order=out_tensor.order.value).new_chunk(
                            [target_l, s], shape=target_l.shape, order=out_tensor.order)
                    u = upper_chunks[j, j]
                    a_transpose = TensorTranspose(dtype=u.dtype, sparse=op.sparse).new_chunk([u], shape=u.shape)
                    target_transpose = TensorTranspose(dtype=target_l.dtype, sparse=op.sparse).new_chunk(
                        [target_l], shape=target_l.shape)
                    lower_permuted_chunk = TensorSolveTriangular(
                        lower=True, dtype=L.dtype, strict=False, sparse=op.sparse).new_chunk(
                        [a_transpose, target_transpose], shape=target_l.shape, index=(i, j),
                        order=out_tensor.order)
                    lower_transpose = TensorTranspose(dtype=lower_permuted_chunk.dtype, sparse=op.sparse).new_chunk(
                        [lower_permuted_chunk], shape=lower_permuted_chunk.shape, index=lower_permuted_chunk.index)
                    l_permuted_chunks[lower_permuted_chunk.index] = lower_transpose

        new_op = op.copy()
        kws = [
            {'chunks': list(p_chunks.values()), 'nsplits': in_tensor.nsplits,
             'dtype': P.dtype, 'shape': P.shape, 'order': P.order},
            {'chunks': list(lower_chunks.values()), 'nsplits': in_tensor.nsplits,
             'dtype': L.dtype, 'shape': L.shape, 'order': L.order},
            {'chunks': list(upper_chunks.values()), 'nsplits': in_tensor.nsplits,
             'dtype': U.dtype, 'shape': U.shape, 'order': U.order}
        ]
        if raw_in_tensor.shape[0] == raw_in_tensor.shape[1]:
            return new_op.new_tensors(op.inputs, kws=kws)

        p, l_, u = new_op.new_tensors(op.inputs, kws=kws)
        if raw_in_tensor.shape[0] > raw_in_tensor.shape[1]:
            l_ = l_[:, :raw_in_tensor.shape[1]]._inplace_tile()
            u = u[:raw_in_tensor.shape[1], :raw_in_tensor.shape[1]]._inplace_tile()
        else:
            p = p[:raw_in_tensor.shape[0], :raw_in_tensor.shape[0]]._inplace_tile()
            l_ = l_[:raw_in_tensor.shape[0], :raw_in_tensor.shape[0]]._inplace_tile()
            u = u[:raw_in_tensor.shape[0], :]._inplace_tile()
        kws = [
            {'chunks': p.chunks, 'nsplits': p.nsplits, 'dtype': P.dtype,
             'shape': p.shape, 'order': p.order},
            {'chunks': l_.chunks, 'nsplits': l_.nsplits, 'dtype': l_.dtype,
             'shape': l_.shape, 'order': l_.order},
            {'chunks': u.chunks, 'nsplits': u.nsplits, 'dtype': u.dtype,
             'shape': u.shape, 'order': u.order}
        ]
        return new_op.new_tensors(op.inputs, kws=kws)

    @classmethod
    def execute(cls, ctx, op):
        (a,), device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True)

        with device(device_id):
            if xp is np:
                import scipy.linalg

                p, l, u = scipy.linalg.lu(a)
            elif is_sparse_module(xp):
                p, l, u = xp.lu(a)
            else:
                raise NotImplementedError
            pc, lc, uc = op.outputs

            ctx[pc.key] = p
            ctx[lc.key] = l
            ctx[uc.key] = u


def lu(a):
    """
    LU decomposition

    The decomposition is::
        A = P L U
    where P is a permutation matrix, L lower triangular with unit diagonal elements,
    and U upper triangular.

    Parameters
    ----------
    a : (M, N) array_like
        Array to decompose

    Returns
    -------
    p : (M, M) ndarray
        Permutation matrix
    l : (M, K) ndarray
        Lower triangular or trapezoidal matrix with unit diagonal.
        K = min(M, N)
    u : (K, N) ndarray
        Upper triangular or trapezoidal matrix

    Examples
    --------
    >>> import mars.tensor as mt

    >>> A = mt.array([[1,2],[2,3]])
    >>> A.execute()
    array([[ 1,  2],
           [ 2,  3]])
    >>> P, L, U = mt.linalg.lu(A)
    >>> P.execute()
    array([[ 0,  1],
           [ 1,  0]])
    >>> L.execute()
    array([[ 1,  0],
           [ 0.5,  1]])
    >>> U.execute()
    array([[ 2,  3],
           [ 0,  0.5]])
    >>> mt.dot(P.dot(L), U).execute() # verify that PL * U = A
    array([[ 1,  2],
           [ 2,  3]])

    """
    op = TensorLU(sparse=a.issparse())
    return op(a)



# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import logging
import random
import time
import uuid
from urllib.parse import urlparse

from ...session import new_session
from ..utils import wait_services_ready
from .config import NamespaceConfig, RoleConfig, RoleBindingConfig, ServiceConfig, \
    MarsSchedulersConfig, MarsWorkersConfig, MarsWebsConfig

try:
    from kubernetes.client.rest import ApiException as K8SApiException
except ImportError:  # pragma: no cover
    K8SApiException = None

logger = logging.getLogger(__name__)


class KubernetesClusterClient:
    def __init__(self, cluster):
        self._cluster = cluster
        self._endpoint = None
        self._session = None

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def namespace(self):
        return self._cluster.namespace

    @property
    def session(self):
        return self._session

    def start(self):
        self._endpoint = self._cluster.start()
        self._session = new_session(self._endpoint).as_default()

    def stop(self, wait=False, timeout=0):
        self._cluster.stop(wait=wait, timeout=timeout)

    def rescale_workers(self, new_scale, min_workers=None, wait=True, timeout=None):
        self._session._sess.rescale_workers(
            new_scale, min_workers=min_workers, wait=wait, timeout=timeout)


class KubernetesCluster:
    _scheduler_config_cls = MarsSchedulersConfig
    _worker_config_cls = MarsWorkersConfig
    _web_config_cls = MarsWebsConfig
    _default_service_port = 7103

    def __init__(self, kube_api_client=None, image=None, namespace=None,
                 scheduler_num=1, scheduler_cpu=None, scheduler_mem=None,
                 scheduler_mem_limit_ratio=None,
                 worker_num=1, worker_cpu=None, worker_mem=None,
                 worker_spill_paths=None, worker_cache_mem=None, min_worker_num=None,
                 worker_min_cache_mem=None, worker_mem_limit_ratio=None,
                 web_num=1, web_cpu=None, web_mem=None, web_mem_limit_ratio=None,
                 service_type=None, timeout=None, **kwargs):
        from kubernetes import client as kube_client

        self._api_client = kube_api_client
        self._core_api = kube_client.CoreV1Api(kube_api_client)
        self._rbac_api = kube_client.RbacAuthorizationV1Api(kube_api_client)

        self._namespace = namespace
        self._image = image
        self._timeout = timeout
        self._service_type = service_type or 'NodePort'
        self._extra_volumes = kwargs.pop('extra_volumes', ())
        self._pre_stop_command = kwargs.pop('pre_stop_command', None)
        self._log_when_fail = kwargs.pop('log_when_fail', False)

        extra_modules = kwargs.pop('extra_modules', None) or []
        extra_modules = extra_modules.split(',') if isinstance(extra_modules, str) \
            else extra_modules
        extra_envs = kwargs.pop('extra_env', None) or dict()
        extra_labels = kwargs.pop('extra_labels', None) or dict()
        service_port = kwargs.pop('service_port', None) or self._default_service_port

        def _override_modules(updates):
            modules = set(extra_modules)
            updates = updates.split(',') if isinstance(updates, str) \
                else updates
            modules.update(updates)
            return sorted(modules)

        def _override_dict(d, updates):
            updates = updates or dict()
            ret = d.copy()
            ret.update(updates)
            return ret

        _override_envs = functools.partial(_override_dict, extra_envs)
        _override_labels = functools.partial(_override_dict, extra_labels)

        self._scheduler_num = scheduler_num
        self._scheduler_cpu = scheduler_cpu
        self._scheduler_mem = scheduler_mem
        self._scheduler_mem_limit_ratio = scheduler_mem_limit_ratio
        self._scheduler_extra_modules = _override_modules(kwargs.pop('scheduler_extra_modules', []))
        self._scheduler_extra_env = _override_envs(kwargs.pop('scheduler_extra_env', None))
        self._scheduler_extra_labels = _override_labels(kwargs.pop('scheduler_extra_labels', None))
        self._scheduler_service_port = kwargs.pop('scheduler_service_port', None) or service_port

        self._worker_num = worker_num
        self._worker_cpu = worker_cpu
        self._worker_mem = worker_mem
        self._worker_mem_limit_ratio = worker_mem_limit_ratio
        self._worker_spill_paths = worker_spill_paths
        self._worker_cache_mem = worker_cache_mem
        self._worker_min_cache_men = worker_min_cache_mem
        self._min_worker_num = min_worker_num
        self._worker_extra_modules = _override_modules(kwargs.pop('worker_extra_modules', []))
        self._worker_extra_env = _override_envs(kwargs.pop('worker_extra_env', None))
        self._worker_extra_labels = _override_labels(kwargs.pop('worker_extra_labels', None))
        self._worker_service_port = kwargs.pop('worker_service_port', None) or service_port

        self._web_num = web_num
        self._web_cpu = web_cpu
        self._web_mem = web_mem
        self._web_mem_limit_ratio = web_mem_limit_ratio
        self._web_extra_modules = _override_modules(kwargs.pop('web_extra_modules', []))
        self._web_extra_env = _override_envs(kwargs.pop('web_extra_env', None))
        self._web_extra_labels = _override_labels(kwargs.pop('web_extra_labels', None))
        self._web_service_port = kwargs.pop('web_service_port', None) or service_port

    @property
    def namespace(self):
        return self._namespace

    def _get_free_namespace(self):
        while True:
            namespace = 'mars-ns-' + str(uuid.uuid4().hex)
            try:
                self._core_api.read_namespace(namespace)
            except K8SApiException as ex:
                if ex.status != 404:  # pragma: no cover
                    raise
                return namespace

    def _create_kube_service(self):
        if self._service_type != 'NodePort':  # pragma: no cover
            raise NotImplementedError(f'Service type {self._service_type} not supported')

        service_config = ServiceConfig(
            'marsservice', service_type='NodePort', port=self._web_service_port,
            selector={'mars/service-type': MarsWebsConfig.rc_name},
        )
        self._core_api.create_namespaced_service(self._namespace, service_config.build())

        time.sleep(1)

        svc_data = self._core_api.read_namespaced_service('marsservice', self._namespace).to_dict()
        port = svc_data['spec']['ports'][0]['node_port']

        web_pods = self._core_api.list_namespaced_pod(
            self._namespace, label_selector='mars/service-type=' + MarsWebsConfig.rc_name).to_dict()
        host_ip = random.choice(web_pods['items'])['status']['host_ip']

        # docker desktop use a VM to hold docker processes, hence
        # we need to use API address instead
        desktop_nodes = self._core_api.list_node(field_selector='metadata.name=docker-desktop').to_dict()
        if desktop_nodes['items']:  # pragma: no cover
            addresses = set(
                addr_info['address']
                for addr_info in desktop_nodes['items'][0].get('status', {}).get('addresses', ())
            )
            if host_ip in addresses:
                host_ip = urlparse(self._core_api.api_client.configuration.host).netloc.split(':', 1)[0]
        return f'http://{host_ip}:{port}'

    def _get_ready_pod_count(self, label_selector):
        query = self._core_api.list_namespaced_pod(
            namespace=self._namespace, label_selector=label_selector).to_dict()
        cnt = 0
        for el in query['items']:
            if el['status']['phase'] in ('Error', 'Failed'):
                logger.warning('Error in starting pod, message: %s', el['status']['message'])
                continue
            if 'status' not in el or 'conditions' not in el['status']:
                cnt += 1
            elif any(cond['type'] == 'Ready' and cond['status'] == 'True'
                     for cond in el['status'].get('conditions') or ()):
                cnt += 1
        return cnt

    def _create_namespace(self):
        if self._namespace is None:
            namespace = self._namespace = self._get_free_namespace()
        else:
            namespace = self._namespace

        self._core_api.create_namespace(NamespaceConfig(namespace).build())

    def _create_roles_and_bindings(self):
        # create role and binding
        role_config = RoleConfig('mars-pod-operator', self._namespace, api_groups='',
                                 resources='pods,replicationcontrollers/scale',
                                 verbs='get,watch,list,patch')
        self._rbac_api.create_namespaced_role(self._namespace, role_config.build())
        role_binding_config = RoleBindingConfig(
            'mars-pod-operator-binding', self._namespace, 'mars-pod-operator', 'default')
        self._rbac_api.create_namespaced_role_binding(self._namespace, role_binding_config.build())

    def _create_schedulers(self):
        schedulers_config = self._scheduler_config_cls(
            self._scheduler_num, image=self._image, cpu=self._scheduler_cpu,
            memory=self._scheduler_mem, memory_limit_ratio=self._scheduler_mem_limit_ratio,
            modules=self._scheduler_extra_modules, volumes=self._extra_volumes,
            service_port=self._scheduler_service_port,
            pre_stop_command=self._pre_stop_command,
        )
        schedulers_config.add_simple_envs(self._scheduler_extra_env)
        schedulers_config.add_labels(self._scheduler_extra_labels)
        self._core_api.create_namespaced_replication_controller(
            self._namespace, schedulers_config.build())

    def _create_workers(self):
        workers_config = self._worker_config_cls(
            self._worker_num, image=self._image, cpu=self._worker_cpu,
            memory=self._worker_mem, memory_limit_ratio=self._worker_mem_limit_ratio,
            spill_volumes=self._worker_spill_paths,
            modules=self._worker_extra_modules, volumes=self._extra_volumes,
            worker_cache_mem=self._worker_cache_mem,
            min_cache_mem=self._worker_min_cache_men,
            service_port=self._worker_service_port,
            pre_stop_command=self._pre_stop_command,
        )
        workers_config.add_simple_envs(self._worker_extra_env)
        workers_config.add_labels(self._worker_extra_labels)
        self._core_api.create_namespaced_replication_controller(
            self._namespace, workers_config.build())

    def _create_webs(self):
        webs_config = self._web_config_cls(
            self._web_num, image=self._image, cpu=self._web_cpu, memory=self._web_mem,
            memory_limit_ratio=self._web_mem_limit_ratio, modules=self._web_extra_modules,
            volumes=self._extra_volumes, service_port=self._web_service_port,
            pre_stop_command=self._pre_stop_command,
        )
        webs_config.add_simple_envs(self._web_extra_env)
        webs_config.add_labels(self._web_extra_labels)
        self._core_api.create_namespaced_replication_controller(
            self._namespace, webs_config.build())

    def _create_services(self):
        self._create_schedulers()
        self._create_workers()
        self._create_webs()

    def _wait_services_ready(self):
        min_worker_num = int(self._min_worker_num or self._worker_num)
        limits = [self._scheduler_num, min_worker_num, self._web_num]
        selectors = [
            'mars/service-type=' + MarsSchedulersConfig.rc_name,
            'mars/service-type=' + MarsWorkersConfig.rc_name,
            'mars/service-type=' + MarsWebsConfig.rc_name,
            ]
        logger.debug('Start waiting pods to be ready')
        wait_services_ready(selectors, limits,
                            lambda sel: self._get_ready_pod_count(sel),
                            timeout=self._timeout)
        logger.info('All service pods ready.')

    def _load_cluster_logs(self):
        log_dict = dict()
        pod_items = self._core_api.list_namespaced_pod(self._namespace).to_dict()
        for item in pod_items['items']:
            log_dict[item['metadata']['name']] = self._core_api.read_namespaced_pod_log(
                name=item['metadata']['name'], namespace=self._namespace)
        return log_dict

    def start(self):
        try:
            self._create_namespace()
            self._create_roles_and_bindings()

            self._create_services()

            self._wait_services_ready()
            return self._create_kube_service()
        except:  # noqa: E722
            if self._log_when_fail:  # pargma: no cover
                logger.error('Error when creating cluster')
                for name, log in self._load_cluster_logs().items():
                    logger.error('Error logs for %s:\n%s', name, log)
            self.stop()
            raise

    def stop(self, wait=False, timeout=0):
        from kubernetes.client import CoreV1Api
        api = CoreV1Api(self._api_client)
        api.delete_namespace(self._namespace)
        if wait:
            start_time = time.time()
            while True:
                try:
                    api.read_namespace(self._namespace)
                except K8SApiException as ex:
                    if ex.status != 404:  # pragma: no cover
                        raise
                    break
                else:
                    time.sleep(1)
                    if timeout and time.time() - start_time > timeout:  # pragma: no cover
                        raise TimeoutError


def new_cluster(kube_api_client=None, image=None, scheduler_num=1, scheduler_cpu=None,
                scheduler_mem=None, worker_num=1, worker_cpu=None, worker_mem=None,
                worker_spill_paths=None, worker_cache_mem=None, min_worker_num=None,
                web_num=1, web_cpu=None, web_mem=None, service_type=None,
                timeout=None, **kwargs):
    """
    :param kube_api_client: Kubernetes API client, can be created with ``new_client_from_config``
    :param image: Docker image to use, ``marsproject/mars:<mars version>`` by default
    :param scheduler_num: Number of schedulers in the cluster, 1 by default
    :param scheduler_cpu: Number of CPUs for every scheduler
    :param scheduler_mem: Memory size for every scheduler
    :param worker_num: Number of workers in the cluster, 1 by default
    :param worker_cpu: Number of CPUs for every worker
    :param worker_mem: Memory size for every worker
    :param worker_spill_paths: Spill paths for worker pods on hosts
    :param worker_cache_mem: Size or ratio of cache memory for every worker
    :param min_worker_num: Minimal ready workers
    :param web_num: Number of web services in the cluster, 1 by default
    :param web_cpu: Number of CPUs for every web service
    :param web_mem: Memory size for every web service
    :param service_type: Type of Kubernetes Service, currently only ``NodePort`` supported
    :param timeout: Timeout when creating clusters
    """
    cluster_cls = kwargs.pop('cluster_cls', KubernetesCluster)
    cluster = cluster_cls(
        kube_api_client, image=image, scheduler_num=scheduler_num, scheduler_cpu=scheduler_cpu,
        scheduler_mem=scheduler_mem, worker_num=worker_num, worker_cpu=worker_cpu,
        worker_mem=worker_mem, worker_spill_paths=worker_spill_paths, worker_cache_mem=worker_cache_mem,
        min_worker_num=min_worker_num, web_num=web_num, web_cpu=web_cpu, web_mem=web_mem,
        service_type=service_type, timeout=timeout, **kwargs)
    client = KubernetesClusterClient(cluster)
    client.start()
    return client



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from ..actors import new_client
from .session import SessionActor
from .chunkmeta import ChunkMetaClient
from .utils import SchedulerClusterInfoActor


class MetaAPI:
    def __init__(self, actor_ctx=None, scheduler_endpoint=None):
        self._actor_ctx = actor_ctx or new_client()
        self._cluster_info = self._actor_ctx.actor_ref(
            SchedulerClusterInfoActor.default_uid(), address=scheduler_endpoint)
        self._chunk_meta_client = ChunkMetaClient(self._actor_ctx, self._cluster_info)

    def _get_session_ref(self, session_id):
        session_uid = SessionActor.gen_uid(session_id)
        session_ref = self._actor_ctx.actor_ref(session_uid, address=self._cluster_info.get_scheduler(session_uid))
        return session_ref

    def get_tileable_key_by_name(self, session_id, name: str):
        sess_ref = self._get_session_ref(session_id)
        return sess_ref.get_tileable_key(name)

    def get_tileable_metas(self, session_id, tileable_keys, filter_fields: List[str] = None) -> List:
        session_ref = self._get_session_ref(session_id)
        graph_ref = self._actor_ctx.actor_ref(session_ref.get_graph_ref_by_tileable_key(tileable_keys[0]))
        return graph_ref.get_tileable_metas(tileable_keys, filter_fields=filter_fields)

    def get_chunk_metas(self, session_id, chunk_keys, filter_fields: List[str] = None) -> List:
        return self._chunk_meta_client.batch_get_chunk_meta(
            session_id, chunk_keys, filter_fields=filter_fields)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .fillna import fillna, ffill, bfill, index_fillna
from .checkna import isna, notna, isnull, notnull
from .dropna import df_dropna, series_dropna, index_dropna
from .replace import df_replace, series_replace


def _install():
    from ..core import DATAFRAME_TYPE, SERIES_TYPE, INDEX_TYPE

    for cls in DATAFRAME_TYPE + SERIES_TYPE:
        setattr(cls, 'fillna', fillna)
        setattr(cls, 'ffill', ffill)
        setattr(cls, 'pad', ffill)
        setattr(cls, 'backfill', bfill)
        setattr(cls, 'bfill', bfill)
        setattr(cls, 'isna', isna)
        setattr(cls, 'isnull', isnull)
        setattr(cls, 'notna', notna)
        setattr(cls, 'notnull', notnull)

    for cls in DATAFRAME_TYPE:
        setattr(cls, 'dropna', df_dropna)
        setattr(cls, 'replace', df_replace)

    for cls in SERIES_TYPE:
        setattr(cls, 'dropna', series_dropna)
        setattr(cls, 'replace', series_replace)

    for cls in INDEX_TYPE:
        setattr(cls, 'fillna', index_fillna)
        setattr(cls, 'dropna', index_dropna)
        setattr(cls, 'isna', isna)
        setattr(cls, 'notna', notna)


_install()
del _install



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...utils import classproperty


_tensor_op_to_df_op = dict()


def register_tensor_ufunc(op):
    _tensor_op_to_df_op[op.tensor_op_type] = op


def get_tensor_ufunc_implementation(tensor_op):
    if tensor_op in _tensor_op_to_df_op:
        return _tensor_op_to_df_op[tensor_op]


class TensorUfuncMixin:
    __slots__ = ()

    @classproperty
    def tensor_op_type(self):
        raise NotImplementedError

    @classmethod
    def ufunc_call(cls, tensor_op, inputs, out, where, **kw):
        if out is not None:
            return NotImplemented
        if where is not None:
            raise NotImplementedError

        try:
            op = _tensor_op_to_df_op[tensor_op](**kw)
            return op(*inputs)
        except (KeyError, TypeError):
            return NotImplemented


def _tensor_ufunc(_, tensor_op, inputs, out, where, **kw):
    op = get_tensor_ufunc_implementation(tensor_op)
    if op is not None:
        return op.ufunc_call(tensor_op, inputs, out, where, **kw)
    return NotImplemented



# cython: profile=False, boundscheck=False, wraparound=False, cdivision=True
#
# Author: Andreas Mueller
#
# Licence: BSD 3 clause

# TODO: We still need to use ndarrays instead of typed memoryviews when using
# fused types and when the array may be read-only (for instance when it's
# provided by the user). This is fixed in cython > 0.3.

import numpy as np
cimport numpy as np
cimport cython
from cython cimport floating
from sklearn.utils.extmath import row_norms

from ._k_means_fast cimport _euclidean_dense_dense
from ._k_means_fast cimport _euclidean_sparse_dense


np.import_array()


def init_bounds_dense(
        np.ndarray[floating, ndim=2, mode='c'] X,  # IN
        floating[:, ::1] centers,                  # IN
        floating[:, ::1] center_half_distances,    # IN
        int[::1] labels,                           # OUT
        floating[::1] upper_bounds,                # OUT
        floating[:, ::1] lower_bounds):            # OUT
    """Initialize upper and lower bounds for each sample for dense input data.

    Given X, centers and the pairwise distances divided by 2.0 between the
    centers this calculates the upper bounds and lower bounds for each sample.
    The upper bound for each sample is set to the distance between the sample
    and the closest center.

    The lower bound for each sample is a one-dimensional array of n_clusters.
    For each sample i assume that the previously assigned cluster is c1 and the
    previous closest distance is dist, for a new cluster c2, the
    lower_bound[i][c2] is set to distance between the sample and this new
    cluster, if and only if dist > center_half_distances[c1][c2]. This prevents
    computation of unnecessary distances for each sample to the clusters that
    it is unlikely to be assigned to.

    Parameters
    ----------
    X : ndarray of shape (n_samples, n_features), dtype=floating
        The input data.

    centers : ndarray of shape (n_clusters, n_features), dtype=floating
        The cluster centers.

    center_half_distances : ndarray of shape (n_clusters, n_clusters), \
            dtype=floating
        The half of the distance between any 2 clusters centers.

    labels : ndarray of shape(n_samples), dtype=int
        The label for each sample. This array is modified in place.

    upper_bounds : ndarray of shape(n_samples,), dtype=floating
        The upper bound on the distance between each sample and its closest
        cluster center. This array is modified in place.

    lower_bounds : ndarray, of shape(n_samples, n_clusters), dtype=floating
        The lower bound on the distance between each sample and each cluster
        center. This array is modified in place.
    """
    cdef:
        int n_samples = X.shape[0]
        int n_clusters = centers.shape[0]
        int n_features = X.shape[1]

        floating min_dist, dist
        int best_cluster, i, j

    for i in range(n_samples):
        best_cluster = 0
        min_dist = _euclidean_dense_dense(&X[i, 0], &centers[0, 0],
                                          n_features, False)
        lower_bounds[i, 0] = min_dist
        for j in range(1, n_clusters):
            if min_dist > center_half_distances[best_cluster, j]:
                dist = _euclidean_dense_dense(&X[i, 0], &centers[j, 0],
                                              n_features, False)
                lower_bounds[i, j] = dist
                if dist < min_dist:
                    min_dist = dist
                    best_cluster = j
        labels[i] = best_cluster
        upper_bounds[i] = min_dist


def init_bounds_sparse(
        X,                                       # IN
        floating[:, ::1] centers,                # IN
        floating[:, ::1] center_half_distances,  # IN
        int[::1] labels,                         # OUT
        floating[::1] upper_bounds,              # OUT
        floating[:, ::1] lower_bounds):          # OUT
    """Initialize upper and lower bounds for each sample for sparse input data.

    Given X, centers and the pairwise distances divided by 2.0 between the
    centers this calculates the upper bounds and lower bounds for each sample.
    The upper bound for each sample is set to the distance between the sample
    and the closest center.

    The lower bound for each sample is a one-dimensional array of n_clusters.
    For each sample i assume that the previously assigned cluster is c1 and the
    previous closest distance is dist, for a new cluster c2, the
    lower_bound[i][c2] is set to distance between the sample and this new
    cluster, if and only if dist > center_half_distances[c1][c2]. This prevents
    computation of unnecessary distances for each sample to the clusters that
    it is unlikely to be assigned to.

    Parameters
    ----------
    X : sparse matrix of shape (n_samples, n_features), dtype=floating
        The input data. Must be in CSR format.

    centers : ndarray of shape (n_clusters, n_features), dtype=floating
        The cluster centers.

    center_half_distances : ndarray of shape (n_clusters, n_clusters), \
            dtype=floating
        The half of the distance between any 2 clusters centers.

    labels : ndarray of shape(n_samples), dtype=int
        The label for each sample. This array is modified in place.

    upper_bounds : ndarray of shape(n_samples,), dtype=floating
        The upper bound on the distance between each sample and its closest
        cluster center. This array is modified in place.

    lower_bounds : ndarray of shape(n_samples, n_clusters), dtype=floating
        The lower bound on the distance between each sample and each cluster
        center. This array is modified in place.
    """
    cdef:
        int n_samples = X.shape[0]
        int n_clusters = centers.shape[0]
        int n_features = X.shape[1]

        floating[::1] X_data = X.data
        int[::1] X_indices = X.indices
        int[::1] X_indptr = X.indptr

        floating min_dist, dist
        int best_cluster, i, j

        floating[::1] centers_squared_norms = row_norms(centers, squared=True)

    for i in range(n_samples):
        best_cluster = 0
        min_dist = _euclidean_sparse_dense(
            X_data[X_indptr[i]: X_indptr[i + 1]],
            X_indices[X_indptr[i]: X_indptr[i + 1]],
            centers[0], centers_squared_norms[0], False)

        lower_bounds[i, 0] = min_dist
        for j in range(1, n_clusters):
            if min_dist > center_half_distances[best_cluster, j]:
                dist = _euclidean_sparse_dense(
                    X_data[X_indptr[i]: X_indptr[i + 1]],
                    X_indices[X_indptr[i]: X_indptr[i + 1]],
                    centers[j], centers_squared_norms[j], False)
                lower_bounds[i, j] = dist
                if dist < min_dist:
                    min_dist = dist
                    best_cluster = j
        labels[i] = best_cluster
        upper_bounds[i] = min_dist


def update_chunk_dense(
        np.ndarray[floating, ndim=2, mode='c'] X,  # IN
        floating[::1] sample_weight,               # IN
        floating[:, ::1] centers_old,              # IN
        floating[:, ::1] center_half_distances,    # IN
        floating[::1] distance_next_center,        # IN
        int[::1] labels,                           # INOUT
        floating[::1] upper_bounds,                # INOUT
        floating[:, ::1] lower_bounds,             # INOUT
        floating[:, ::1] centers_new,              # OUT
        floating[::1] weight_in_clusters,          # OUT
        bint update_centers=True):
    return _update_chunk_dense(&X[0, 0], sample_weight, centers_old,
                               center_half_distances,
                               distance_next_center, labels,
                               upper_bounds, lower_bounds,
                               &centers_new[0, 0], &weight_in_clusters[0],
                               update_centers)


cdef void _update_chunk_dense(
        floating *X,                             # IN
        # expecting C alinged 2D array. XXX: Can be
        # replaced by const memoryview when cython min
        # version is >= 0.3
        floating[::1] sample_weight,             # IN
        floating[:, ::1] centers_old,            # IN
        floating[:, ::1] center_half_distances,  # IN
        floating[::1] distance_next_center,      # IN
        int[::1] labels,                         # INOUT
        floating[::1] upper_bounds,              # INOUT
        floating[:, ::1] lower_bounds,           # INOUT
        floating *centers_new,                   # OUT
        floating *weight_in_clusters,            # OUT
        bint update_centers) nogil:
    """K-means combined EM step for one dense data chunk.

    Compute the partial contribution of a single data chunk to the labels and
    centers.
    """
    cdef:
        int n_samples = labels.shape[0]
        int n_clusters = centers_old.shape[0]
        int n_features = centers_old.shape[1]

        floating upper_bound, distance
        int i, j, k, label

    for i in range(n_samples):
        upper_bound = upper_bounds[i]
        bounds_tight = 0
        label = labels[i]

        # Next center is not far away from the currently assigned center.
        # Sample might need to be assigned to another center.
        if not distance_next_center[label] >= upper_bound:

            for j in range(n_clusters):

                # If this holds, then center_index is a good candidate for the
                # sample to be relabelled, and we need to confirm this by
                # recomputing the upper and lower bounds.
                if (j != label
                    and (upper_bound > lower_bounds[i, j])
                    and (upper_bound > center_half_distances[label, j])):

                    # Recompute upper bound by calculating the actual distance
                    # between the sample and its current assigned center.
                    if not bounds_tight:
                        upper_bound = _euclidean_dense_dense(
                            X + i * n_features, &centers_old[label, 0], n_features, False)
                        lower_bounds[i, label] = upper_bound
                        bounds_tight = 1

                    # If the condition still holds, then compute the actual
                    # distance between the sample and center. If this is less
                    # than the previous distance, reassign label.
                    if (upper_bound > lower_bounds[i, j]
                        or (upper_bound > center_half_distances[label, j])):

                        distance = _euclidean_dense_dense(
                            X + i * n_features, &centers_old[j, 0], n_features, False)
                        lower_bounds[i, j] = distance
                        if distance < upper_bound:
                            label = j
                            upper_bound = distance

            labels[i] = label
            upper_bounds[i] = upper_bound

        if update_centers:
            weight_in_clusters[label] += sample_weight[i]
            for k in range(n_features):
                centers_new[label * n_features + k] += X[i * n_features + k] * sample_weight[i]


def update_chunk_sparse(
        X,                                         # IN
        floating[::1] sample_weight,               # IN
        floating[:, ::1] centers_old,              # IN
        floating[:, ::1] center_half_distances,    # IN
        floating[::1] distance_next_center,        # IN
        int[::1] labels,                           # INOUT
        floating[::1] upper_bounds,                # INOUT
        floating[:, ::1] lower_bounds,             # INOUT
        floating[:, ::1] centers_new,              # OUT
        floating[::1] weight_in_clusters,          # OUT
        bint update_centers=True):
    cdef:
        floating[::1] X_data = X.data
        int[::1] X_indices = X.indices
        int[::1] X_indptr = X.indptr

        floating[::1] centers_squared_norms = row_norms(centers_old, squared=True)

    return _update_chunk_sparse(
        X_data, X_indices, X_indptr, sample_weight, centers_old,
        centers_squared_norms, center_half_distances,
        distance_next_center, labels, upper_bounds, lower_bounds,
        &centers_new[0, 0], &weight_in_clusters[0], update_centers
    )


cdef void _update_chunk_sparse(
        floating[::1] X_data,                    # IN
        int[::1] X_indices,                      # IN
        int[::1] X_indptr,                       # IN
        floating[::1] sample_weight,             # IN
        floating[:, ::1] centers_old,            # IN
        floating[::1] centers_squared_norms,     # IN
        floating[:, ::1] center_half_distances,  # IN
        floating[::1] distance_next_center,      # IN
        int[::1] labels,                         # INOUT
        floating[::1] upper_bounds,              # INOUT
        floating[:, ::1] lower_bounds,           # INOUT
        floating *centers_new,                   # OUT
        floating *weight_in_clusters,            # OUT
        bint update_centers) nogil:
    """K-means combined EM step for one sparse data chunk.

    Compute the partial contribution of a single data chunk to the labels and
    centers.
    """
    cdef:
        int n_samples = labels.shape[0]
        int n_clusters = centers_old.shape[0]
        int n_features = centers_old.shape[1]

        floating upper_bound, distance
        int i, j, k, label
        int s = X_indptr[0]

    for i in range(n_samples):
        upper_bound = upper_bounds[i]
        bounds_tight = 0
        label = labels[i]

        # Next center is not far away from the currently assigned center.
        # Sample might need to be assigned to another center.
        if not distance_next_center[label] >= upper_bound:

            for j in range(n_clusters):

                # If this holds, then center_index is a good candidate for the
                # sample to be relabelled, and we need to confirm this by
                # recomputing the upper and lower bounds.
                if (j != label
                    and (upper_bound > lower_bounds[i, j])
                    and (upper_bound > center_half_distances[label, j])):

                    # Recompute upper bound by calculating the actual distance
                    # between the sample and its current assigned center.
                    if not bounds_tight:
                        upper_bound = _euclidean_sparse_dense(
                            X_data[X_indptr[i] - s: X_indptr[i + 1] - s],
                            X_indices[X_indptr[i] - s: X_indptr[i + 1] - s],
                            centers_old[label], centers_squared_norms[label], False)
                        lower_bounds[i, label] = upper_bound
                        bounds_tight = 1

                    # If the condition still holds, then compute the actual
                    # distance between the sample and center. If this is less
                    # than the previous distance, reassign label.
                    if (upper_bound > lower_bounds[i, j]
                        or (upper_bound > center_half_distances[label, j])):
                        distance = _euclidean_sparse_dense(
                            X_data[X_indptr[i] - s: X_indptr[i + 1] - s],
                            X_indices[X_indptr[i] - s: X_indptr[i + 1] - s],
                            centers_old[j], centers_squared_norms[j], False)
                        lower_bounds[i, j] = distance
                        if distance < upper_bound:
                            label = j
                            upper_bound = distance

            labels[i] = label
            upper_bounds[i] = upper_bound

        if update_centers:
            weight_in_clusters[label] += sample_weight[i]
            for k in range(X_indptr[i] - s, X_indptr[i + 1] - s):
                centers_new[label * n_features + X_indices[k]] += X_data[k] * sample_weight[i]



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random

import pandas as pd

from .... import dataframe as md
from .... import tensor as mt
from . import build_index, search_index


def sample_data(query, sample_count=10000):
    if sample_count > query.shape[0]:
        sample_count = query.shape[0]

    idx = random.sample(range(query.shape[0]), sample_count)
    sample_query = query[idx, :]
    return sample_query, idx


def linear_build_and_search(doc, query, topk, column_number=None, row_number=None, dimension=None, measure_name=None,
                            threads=4):
    if measure_name is None:
        measure_name = "SquaredEuclidean"
    if dimension is None:
        dimension = doc.shape[1]

    index = build_index(tensor=doc, dimension=dimension, column_number=column_number,
                        distance_metric=measure_name,
                        index_builder="LinearBuilder")

    pk_l, distance_l = search_index(tensor=query, threads=threads, row_number=row_number,
                                    distance_metric=measure_name, dimension=dimension,
                                    topk=topk, index=index)

    return pk_l, distance_l


def build_and_search(doc, query, topk, doc_chunk, query_chunk,
                     index_path=None, threads=4, dimension=None, measure_name=None,
                     need_shuffle=False, storage_options=None,
                     index_builder=None, builder_params=None,
                     index_converter=None, index_converter_params=None,
                     index_searcher=None, searcher_params=None,
                     index_reformer=None, index_reformer_params=None):
    if measure_name is None:
        measure_name = "SquaredEuclidean"
    if dimension is None:
        dimension = doc.shape[1]
    if index_builder is None:
        index_builder = "SsgBuilder"
    if builder_params is None:
        builder_params = {}
    if index_converter_params is None:
        index_converter_params = {}
    if index_searcher is None:
        index_searcher = ""
    if searcher_params is None:
        searcher_params = {}
    if index_reformer is None:
        index_reformer = ""
    if index_reformer_params is None:
        index_reformer_params = {}

    doc = md.DataFrame(pd.DataFrame(doc), chunk_size=(doc_chunk, dimension))
    query = mt.tensor(query, chunk_size=(query_chunk, dimension))

    index = build_index(doc, dimension, index_path,
                        need_shuffle, measure_name,
                        index_builder, builder_params,
                        index_converter, index_converter_params,
                        topk, storage_options)

    pk2, distance = search_index(query, topk, index, threads, dimension,
                                 measure_name, index_searcher, searcher_params,
                                 index_reformer, index_reformer_params,
                                 storage_options)

    return pk2, distance



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import itertools
from collections import namedtuple

import numpy as np
import pandas as pd

from ....tensor import tensor as mars_tensor
from ....dataframe import DataFrame as MarsDataFrame, Series as MarsSeries


class LGBMModelType(enum.Enum):
    CLASSIFIER = 0
    REGRESSOR = 1
    RANKER = 2


_model_type_to_model = dict()


def get_model_cls_from_type(model_type: LGBMModelType):
    import lightgbm
    if not _model_type_to_model:
        _model_type_to_model.update({
            LGBMModelType.CLASSIFIER: lightgbm.LGBMClassifier,
            LGBMModelType.REGRESSOR: lightgbm.LGBMRegressor,
            LGBMModelType.RANKER: lightgbm.LGBMRanker,
        })
    return _model_type_to_model[model_type]


TrainTuple = namedtuple('TrainTuple', 'data label sample_weight init_score')


class LGBMScikitLearnBase:
    def __init__(self, *args, **kwargs):
        if args and isinstance(args[0], self._get_lgbm_class()):
            model = args[0]
            super().__init__(**model.get_params())
            self._copy_extra_params(model, self)
        else:
            super().__init__(*args, **kwargs)

    @classmethod
    def _get_lgbm_class(cls):
        try:
            return getattr(cls, '_lgbm_class')
        except AttributeError:
            lgbm_class = next(base for base in cls.__bases__
                              if base.__module__.startswith('lightgbm'))
            cls._lgbm_class = lgbm_class
            return lgbm_class

    @classmethod
    def _get_param_names(cls):
        return cls._get_lgbm_class()._get_param_names()

    @staticmethod
    def _copy_extra_params(source, dest):
        params = source.get_params()
        attributes = source.__dict__
        extra_param_names = set(attributes.keys()).difference(params.keys())
        for name in extra_param_names:
            setattr(dest, name, attributes[name])

    @staticmethod
    def _convert_tileable(obj):
        if isinstance(obj, np.ndarray):
            return mars_tensor(obj)
        elif isinstance(obj, pd.DataFrame):
            return MarsDataFrame(obj)
        elif isinstance(obj, pd.Series):
            return MarsSeries(obj)
        return obj

    @classmethod
    def _wrap_train_tuple(cls, data, label, sample_weight=None, init_score=None):
        data = cls._convert_tileable(data)
        label = cls._convert_tileable(label)
        sample_weight = cls._convert_tileable(sample_weight)
        init_score = cls._convert_tileable(init_score)
        return TrainTuple(data, label, sample_weight, init_score)

    @staticmethod
    def _wrap_eval_tuples(eval_set=None, eval_sample_weight=None, eval_init_score=None):
        if not eval_set:
            return None

        tps = []
        for (data, label), weight, score in zip(eval_set, eval_sample_weight or itertools.repeat(None),
                                                eval_init_score or itertools.repeat(None)):
            tps.append(TrainTuple(data, label, weight, score))
        return tps

    def fit(self, X, y, sample_weight=None, **kwargs):
        raise NotImplementedError

    def predict(self, X, **kwargs):
        raise NotImplementedError

    def predict_proba(self, X, **kwargs):
        raise NotImplementedError

    def load_model(self, model):
        self.set_params(**self.get_params())
        self._copy_extra_params(model, self)
        return self



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes
from ... import tensor as mt
from ...operands import OutputType
from ...serialize import KeyField
from ...tensor.array_utils import as_same_device, device, sparse
from ...tensor.core import TensorOrder
from ...tiles import TilesError
from ...utils import check_chunks_unknown_shape
from ..operands import LearnOperand, LearnOperandMixin
from ._k_means_fast import _inertia_dense, _inertia_sparse, merge_update_chunks


class KMeansInertia(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.KMEANS_INERTIA

    _x = KeyField('x')
    _sample_weight = KeyField('sample_weight')
    _centers = KeyField('centers')
    _labels = KeyField('labels')

    def __init__(self, x=None, sample_weight=None, centers=None,
                 labels=None, output_types=None, **kw):
        super().__init__(_x=x, _sample_weight=sample_weight,
                         _centers=centers, _labels=labels,
                         _output_types=output_types, **kw)
        if self._output_types is None:
            self._output_types = [OutputType.tensor]

    @property
    def x(self):
        return self._x

    @property
    def sample_weight(self):
        return self._sample_weight

    @property
    def centers(self):
        return self._centers

    @property
    def labels(self):
        return self._labels

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        for field in ('_x', '_sample_weight', '_centers', '_labels'):
            if getattr(self, field, None) is not None:
                setattr(self, field, next(inputs_iter))

    def __call__(self):
        params = {
            'shape': (),
            'dtype': np.dtype(float),
            'order': TensorOrder.C_ORDER
        }
        return self.new_tileable([self._x, self._sample_weight,
                                  self._centers, self._labels], kws=[params])

    @classmethod
    def tile(cls, op: "KMeansInertia"):
        check_chunks_unknown_shape(op.inputs, TilesError)
        x = op.x
        x = x.rechunk({1: x.shape[1]})._inplace_tile()
        sample_weight = op.sample_weight.rechunk({0: x.nsplits[0]})._inplace_tile()
        labels = op.labels.rechunk({0: x.nsplits[0]})._inplace_tile()
        centers = op.centers
        centers = centers.rechunk(centers.shape)._inplace_tile()

        out_chunks = []
        for x_chunk, sample_weight_chunk, labels_chunk \
                in zip(x.chunks, sample_weight.chunks, labels.chunks):
            chunk_op = op.copy().reset_key()
            chunk_params = {
                'shape': (1,),
                'dtype': np.dtype(float),
                'order': TensorOrder.C_ORDER,
                'index': x_chunk.index
            }
            out_chunk = chunk_op.new_chunk(
                [x_chunk, sample_weight_chunk, centers.chunks[0],
                 labels_chunk], kws=[chunk_params])
            out_chunks.append(out_chunk)

        new_op = op.copy()
        params = op.outputs[0].params
        params['shape'] = (x.chunk_shape[0],)
        params['chunks'] = out_chunks
        params['nsplits'] = ((1,) * x.chunk_shape[0],)
        out = new_op.new_tileable(op.inputs, kws=[params]).sum()
        out._inplace_tile()
        return [out]

    @classmethod
    def execute(cls, ctx, op):
        (x, sample_weight, centers, labels), device_id, xp = \
            as_same_device([ctx[inp.key] for inp in op.inputs], device=op.device,
                           ret_extra=True, copy_if_not_writeable=True)

        with device(device_id):
            if xp is np:
                method = _inertia_dense
            elif xp is sparse:
                method = _inertia_sparse
            else:  # pragma: no cover
                raise NotImplementedError('Cannot run inertial on GPU')

            result = method(x, sample_weight, centers, labels)
            ctx[op.outputs[0].key] = np.array([result])


def _inertia(X, sample_weight, centers, labels):
    op = KMeansInertia(x=X, sample_weight=sample_weight,
                       centers=centers, labels=labels)
    return op()


def _execute_merge_update(ctx, op):
    inputs, device_id, xp = as_same_device(
        [ctx[inp.key] for inp in op.inputs], op.device,
        ret_extra=True, copy_if_not_writeable=True)
    length = len(inputs) // 2
    assert len(inputs) % 2 == 0
    centers_new_chunks = inputs[:length]
    weight_in_cluster_chunks = inputs[length:]

    with device(device_id):
        weight_in_clusters = np.zeros(op.n_clusters,
                                      dtype=weight_in_cluster_chunks[0].dtype)
        centers_new = np.zeros_like(centers_new_chunks[0])
        n_clusters = op.n_clusters
        n_features = centers_new_chunks[0].shape[1]

        for weight_in_clusters_chunk, centers_new_chunk in \
                zip(weight_in_cluster_chunks, centers_new_chunks):
            merge_update_chunks(n_clusters, n_features,
                                weight_in_clusters, weight_in_clusters_chunk,
                                centers_new, centers_new_chunk)

        # centers new
        ctx[op.outputs[0].key] = centers_new
        # weight_in_clusters
        ctx[op.outputs[1].key] = weight_in_clusters


class KMeansRelocateEmptyClusters(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.KMEANS_RELOCASTE_EMPTY_CLUSTERS

    _empty_clusters = KeyField('empty_clusters')
    _far_x = KeyField('far_x')
    _far_labels = KeyField('far_labels')
    _far_sample_weights = KeyField('far_sample_weight')
    _centers_new = KeyField('centers_new')
    _weight_in_clusters = KeyField('weight_in_clusters')

    def __init__(self, empty_clusters=None, far_x=None, far_labels=None,
                 far_sample_weights=None, centers_new=None,
                 weight_in_clusters=None, output_types=None, **kw):
        super().__init__(_empty_clusters=empty_clusters, _far_x=far_x,
                         _far_labels=far_labels, _far_sample_weights=far_sample_weights,
                         _centers_new=centers_new, _weight_in_clusters=weight_in_clusters,
                         _output_types=output_types, **kw)
        if self._output_types is None:
            self._output_types = [OutputType.tensor] * self.output_limit

    @property
    def empty_clusters(self):
        return self._empty_clusters

    @property
    def far_x(self):
        return self._far_x

    @property
    def far_labels(self):
        return self._far_labels

    @property
    def far_sample_weights(self):
        return self._far_sample_weights

    @property
    def centers_new(self):
        return self._centers_new

    @property
    def weight_in_clusters(self):
        return self._weight_in_clusters

    @property
    def output_limit(self):
        return 2

    @property
    def _input_fields(self):
        return '_empty_clusters', '_far_x', '_far_labels', \
               '_far_sample_weights', '_centers_new', '_weight_in_clusters'

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        for field in self._input_fields:
            ob = getattr(self, field)
            if ob is not None:
                setattr(self, field, next(inputs_iter))

    def __call__(self):
        kws = [
            # centers_new
            self._centers_new.params,
            # weight_in_clusters
            self._weight_in_clusters.params
        ]
        return self.new_tileables(
            [getattr(self, field) for field in self._input_fields], kws=kws)

    @classmethod
    def tile(cls, op: "KMeansRelocateEmptyClusters"):
        empty_clusters = op.empty_clusters.rechunk(
            op.empty_clusters.shape)._inplace_tile()
        far_x = op.far_x.rechunk(op.far_x.shape)._inplace_tile()
        far_labels = op.far_labels.rechunk(op.far_labels.shape)._inplace_tile()
        far_sample_weight = op.far_sample_weights.rechunk(
            op.far_sample_weights.shape)._inplace_tile()
        centers_new = op.centers_new.rechunk(op.centers_new.shape)._inplace_tile()
        weight_in_clusters = op.weight_in_clusters.rechunk(
            op.weight_in_clusters.shape)._inplace_tile()

        chunk_op = op.copy().reset_key()
        out_centers_new_chunk, out_weight_in_clusters_chunk = chunk_op.new_chunks(
            [empty_clusters.chunks[0], far_x.chunks[0], far_labels.chunks[0],
             far_sample_weight.chunks[0], centers_new.chunks[0],
             weight_in_clusters.chunks[0]], kws=[centers_new.chunks[0].params,
                                                 weight_in_clusters.chunks[0].params])

        out_centers_new_params = centers_new.params
        out_centers_new_params['nsplits'] = centers_new.nsplits
        out_centers_new_params['chunks'] = [out_centers_new_chunk]
        out_weight_in_clusters_params = weight_in_clusters.params
        out_weight_in_clusters_params['nsplits'] = weight_in_clusters.nsplits
        out_weight_in_clusters_params['chunks'] = [out_weight_in_clusters_chunk]
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=[out_centers_new_params,
                                                    out_weight_in_clusters_params])

    @classmethod
    def execute(cls, ctx, op):
        (empty_clusters, far_x, far_labels, far_sample_weight,
         center_new, weight_in_clusters), device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs], op.device, ret_extra=True)

        out_centers_new = center_new.copy()
        out_weight_in_clusters = weight_in_clusters.copy()
        del center_new, weight_in_clusters

        n_empty = empty_clusters.shape[0]
        n_features = far_x.shape[1]

        for idx in range(n_empty):
            new_cluster_id = empty_clusters[idx]
            weight = far_sample_weight[idx]
            old_cluster_id = far_labels[idx]

            for k in range(n_features):
                out_centers_new[old_cluster_id, k] -= far_x[idx, k] * weight
                out_centers_new[new_cluster_id, k] = far_x[idx, k] * weight

            out_weight_in_clusters[new_cluster_id] = weight
            out_weight_in_clusters[old_cluster_id] -= weight

        ctx[op.outputs[0].key] = out_centers_new
        ctx[op.outputs[1].key] = out_weight_in_clusters


def _relocate_empty_clusters(X, sample_weight, centers_old, centers_new,
                             weight_in_clusters, labels, to_run=None,
                             session=None, run_kwargs=None):
    to_run = to_run or list()
    empty_clusters = mt.where(mt.equal(weight_in_clusters, 0))[0].astype(mt.int32)
    to_run.append(empty_clusters)

    mt.ExecutableTuple(to_run).execute(session=session, **(run_kwargs or dict()))

    n_empty = empty_clusters.shape[0]

    if n_empty == 0:
        return centers_new, weight_in_clusters

    distances = ((mt.asarray(X) - mt.asarray(centers_old)[labels]) ** 2).sum(axis=1)
    far_from_centers = \
        mt.argpartition(distances, -n_empty)[:-n_empty-1:-1].astype(np.int32)

    far_x = X[far_from_centers]
    far_labels = labels[far_from_centers]
    far_sample_weight = sample_weight[far_from_centers]

    op = KMeansRelocateEmptyClusters(
        empty_clusters=empty_clusters, far_x=far_x, far_labels=far_labels,
        far_sample_weights=far_sample_weight, centers_new=centers_new,
        weight_in_clusters=weight_in_clusters)
    return op()



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...utils import copy_tileables
from ...dataframe.base.value_counts import DataFrameValueCounts
from ...dataframe.datasource.core import HeadOptimizedDataSource
from ...dataframe.indexing.iloc import DataFrameIlocGetItem, SeriesIlocGetItem
from ...dataframe.sort.core import DataFrameSortOperand
from ...dataframe.utils import parse_index
from .core import TileableOptimizeRule, register


class HeadPushDown(TileableOptimizeRule):
    def match(self, node):
        op = node.op
        inputs = op.inputs
        accept_types = (HeadOptimizedDataSource, DataFrameSortOperand, DataFrameValueCounts)
        if len(inputs) == 1 and isinstance(op, (DataFrameIlocGetItem, SeriesIlocGetItem)) and \
                op.can_be_optimized() and isinstance(inputs[0].op, accept_types) and \
                inputs[0] not in self._optimizer_context.result_tileables:
            return True
        return False

    def apply(self, node):
        op = node.op
        inp = op.inputs[0]
        nrows = inp.op.nrows or 0
        head = op.indexes[0].stop

        if inp in self._optimizer_context:
            new_inp = self._optimizer_context[inp]
        else:
            new_inp = copy_tileables([inp])[0].data

        new_inp.op._nrows = max(nrows, head)
        new_inp._key = node.key
        new_inp._shape = (nrows,) + inp.shape[1:]
        pd_index = node.index_value.to_pandas()[:nrows]
        new_inp._index_value = parse_index(pd_index, node)

        self._optimizer_context[node] = new_inp
        return new_inp


register(DataFrameIlocGetItem, HeadPushDown)
register(SeriesIlocGetItem, HeadPushDown)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Integral

import numpy as np

from ... import opcodes as OperandDef
from ...serialize import ValueType, KeyField, ListField, TupleField, Int32Field
from ...core import Base, Entity
from ...operands import OperandStage
from ...utils import get_shuffle_input_keys_idxes
from ..core import TENSOR_TYPE, TensorOrder
from ..utils import split_indexes_into_chunks, calc_pos, filter_inputs
from ..operands import TensorHasInput, TensorOperandMixin, TensorMapReduceOperand
from ..array_utils import get_array_module
from .core import process_index, calc_shape
from .index_lib import TensorIndexesHandler

FANCY_INDEX_TYPES = TENSOR_TYPE + (np.ndarray,)


class TensorIndex(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.INDEX

    _input = KeyField('input')
    _indexes = ListField('indexes')

    def __init__(self, dtype=None, sparse=False, indexes=None, create_view=False, **kw):
        super().__init__(_dtype=dtype, _sparse=sparse, _indexes=indexes,
                         _create_view=create_view, **kw)

    @property
    def indexes(self):
        return self._indexes

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs[1:])
        new_indexes = [next(inputs_iter) if isinstance(index, (Base, Entity)) else index
                       for index in self._indexes]
        self._indexes = new_indexes

    def on_output_modify(self, new_output):
        from .setitem import TensorIndexSetValue

        if self._create_view:
            a = self.input
            op = TensorIndexSetValue(dtype=a.dtype, sparse=a.issparse(),
                                     indexes=self._indexes, value=new_output)
            return op(a, self._indexes, new_output)

    def on_input_modify(self, new_input):
        if self._create_view:
            new_op = self.copy().reset_key()
            new_inputs = [new_input] + self.inputs[1:]
            return new_op.new_tensor(new_inputs, shape=self.outputs[0].shape)

    def __call__(self, a, index, shape, order):
        self._indexes = index
        return self.new_tensor(filter_inputs([a] + list(index)), shape, order=order)

    @classmethod
    def tile(cls, op):
        handler = TensorIndexesHandler()
        return [handler.handle(op)]

    @classmethod
    def execute(cls, ctx, op):
        indexes = tuple(ctx[index.key] if hasattr(index, 'key') else index
                        for index in op.indexes)
        input_ = ctx[op.inputs[0].key]
        xp = get_array_module(input_)
        ret = xp.asarray(input_)[indexes]
        if hasattr(ret, 'astype'):
            ret = ret.astype(
                ret.dtype, order=op.outputs[0].order.value, copy=False)
        ctx[op.outputs[0].key] = ret

    @classmethod
    def estimate_size(cls, ctx, op):
        from mars.core import Base, Entity
        chunk = op.outputs[0]
        shape = chunk.shape
        new_indexes = [index for index in op._indexes if index is not None]

        new_shape = []
        first_fancy_index = False
        for index in new_indexes:
            if isinstance(index, (Base, Entity)):
                if index.dtype != np.bool_:
                    if not first_fancy_index:
                        first_fancy_index = True
                    else:
                        continue
                new_shape.append(ctx[index.key][0] // index.dtype.itemsize)

        rough_shape = []
        idx = 0
        for s in shape:
            if np.isnan(s):
                rough_shape.append(new_shape[idx])
                idx += 1
            else:
                rough_shape.append(s)
        result = int(np.prod(rough_shape) * chunk.dtype.itemsize)
        ctx[chunk.key] = (result, result)


class FancyIndexingDistribute(TensorMapReduceOperand, TensorOperandMixin):
    _op_type_ = OperandDef.FANCY_INDEX_DISTRIBUTE

    _input = KeyField('input')
    _dest_nsplits = TupleField('dest_nsplits', ValueType.tuple(ValueType.uint64))
    _axes = TupleField('axes', ValueType.int32)

    def __init__(self, stage=None, dest_nsplits=None, axes=None, dtype=None, sparse=None,
                 shuffle_key=None, **kw):
        super().__init__(_stage=stage, _dest_nsplits=dest_nsplits, _axes=axes,
                         _dtype=dtype, _sparse=sparse, _shuffle_key=shuffle_key, **kw)

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    @property
    def output_limit(self):
        if self.stage == OperandStage.map:
            return 1
        # return fancy indexes on each axis as well as original position
        return len(self._axes) + 1

    @property
    def dest_nsplits(self):
        return self._dest_nsplits

    @property
    def axes(self):
        return self._axes

    @classmethod
    def _execute_map(cls, ctx, op):
        nsplits = op.dest_nsplits
        axes = op.axes
        fancy_index_nsplits = [nsplits[ax] for ax in axes]
        indexes = ctx[op.inputs[0].key]
        flatten_indexes = indexes.reshape(indexes.shape[0], -1)
        idx_to_fancy_indexes, idx_to_poses = \
            split_indexes_into_chunks(fancy_index_nsplits, flatten_indexes, False)
        for idx in idx_to_fancy_indexes:
            group_key = ','.join(str(i) for i in idx)
            ctx[(op.outputs[0].key, group_key)] = (idx_to_fancy_indexes[idx], idx_to_poses[idx])

    @classmethod
    def _execute_reduce(cls, ctx, op):
        in_chunk = op.inputs[0]

        input_keys, _ = get_shuffle_input_keys_idxes(in_chunk)

        fancy_indexes = []
        poses = []
        shuffle_key = op.shuffle_key
        xp = None
        for input_key in input_keys:
            key = (input_key, shuffle_key)
            fancy_index, pos = ctx[key]
            if xp is None:
                xp = get_array_module(fancy_index)
            if fancy_index.size == 0:
                fancy_index = fancy_index.reshape(len(op.axes), 0)
            fancy_indexes.append(fancy_index)
            poses.append(pos)

        fancy_index = np.hstack(fancy_indexes)
        pos = np.hstack(poses)

        assert len(op.outputs) - 1 == len(fancy_index)
        for out_chunk, axis_fancy_index in zip(op.outputs[:-1], fancy_index):
            ctx[out_chunk.key] = axis_fancy_index
        ctx[op.outputs[-1].key] = np.asarray([len(p) for p in poses]), pos

    @classmethod
    def execute(cls, ctx, op):
        if op.stage == OperandStage.map:
            cls._execute_map(ctx, op)
        else:
            cls._execute_reduce(ctx, op)

    @classmethod
    def estimate_size(cls, ctx, op):
        if op.stage == OperandStage.map:
            fancy_index_size = len(op.axes)
            inp_size = ctx[op.inputs[0].key][0]
            factor = 1 / float(fancy_index_size) + fancy_index_size  # 1/#fancy_index is the poses
            ctx[op.outputs[0].key] = (inp_size * factor,) * 2
        else:
            sum_size = 0
            for shuffle_input in op.inputs[0].inputs or ():
                sum_size += ctx[shuffle_input.key]
            for out_chunk in op.outputs:
                ctx[out_chunk.key] = sum_size, sum_size


class FancyIndexingConcat(TensorMapReduceOperand, TensorOperandMixin):
    _op_type_ = OperandDef.FANCY_INDEX_CONCAT

    _fancy_index_axis = Int32Field('fancy_index_axis')
    _fancy_index_shape = TupleField('fancy_index_shape', ValueType.int64)

    def __init__(self, stage=None, fancy_index_axis=None, fancy_index_shape=None,
                 shuffle_key=None, dtype=None, sparse=None, **kw):
        super().__init__(_stage=stage, _fancy_index_axis=fancy_index_axis,
                         _fancy_index_shape=fancy_index_shape,
                         _shuffle_key=shuffle_key, _dtype=dtype, _sparse=sparse, **kw)

    @property
    def input(self):
        return self._input

    @property
    def fancy_index_axis(self):
        return self._fancy_index_axis

    @property
    def fancy_index_shape(self):
        return self._fancy_index_shape

    @classmethod
    def _execute_map(cls, ctx, op):
        indexed_array = ctx[op.inputs[0].key]
        sizes, pos = ctx[op.inputs[1].key]
        acc_sizes = np.cumsum(sizes)
        fancy_index_axis = op.fancy_index_axis

        for i in range(len(sizes)):
            start = 0 if i == 0 else acc_sizes[i - 1]
            end = acc_sizes[i]
            select = (slice(None),) * fancy_index_axis + (slice(start, end),)
            ctx[(op.outputs[0].key, str(i))] = (indexed_array[select], pos[start: end])

    @classmethod
    def _execute_reduce(cls, ctx, op):
        in_chunk = op.inputs[0]
        input_keys, _ = get_shuffle_input_keys_idxes(in_chunk)
        fancy_index_axis = op.fancy_index_axis
        fancy_index_shape = op.fancy_index_shape

        indexed_arrays = []
        poses = []
        shuffle_key = op.shuffle_key
        for input_key in input_keys:
            index_array, pos = ctx[(input_key, shuffle_key)]
            indexed_arrays.append(index_array)
            poses.append(pos)

        concat_array = get_array_module(indexed_arrays[0]).concatenate(
            indexed_arrays, axis=fancy_index_axis)
        concat_pos = get_array_module(poses[0]).hstack(poses)
        select_pos = calc_pos(fancy_index_shape, concat_pos,
                              xp=get_array_module(poses[0]))
        select = (slice(None),) * fancy_index_axis + (select_pos,)
        ctx[op.outputs[0].key] = concat_array[select]

    @classmethod
    def execute(cls, ctx, op):
        if op.stage == OperandStage.map:
            cls._execute_map(ctx, op)
        else:
            cls._execute_reduce(ctx, op)

    @classmethod
    def estimate_size(cls, ctx, op):
        if op.stage == OperandStage.map:
            input_size = ctx[op.inputs[0].key][0]
            pos_size = ctx[op.inputs[0].key][0]
            ctx[op.outputs[0].key] = input_size + pos_size, input_size + pos_size * 2
        else:
            chunk = op.outputs[0]
            input_sizes = [ctx[c.key][0] for c in op.inputs[0].inputs or ()]
            ctx[chunk.key] = chunk.nbytes, chunk.nbytes + sum(input_sizes)


def _is_bool_index(index_obj):
    return isinstance(index_obj, TENSOR_TYPE) and index_obj.dtype == np.bool_


def _is_fancy_index(index_obj):
    return isinstance(index_obj, FANCY_INDEX_TYPES) and index_obj.dtype != np.bool_


def _is_create_view(index):
    # is view if all of index is slice, int or newaxis
    return all(isinstance(ind, (slice, Integral)) or ind is None for ind in index)


def _calc_order(a, index):
    if a.order == TensorOrder.C_ORDER:
        return TensorOrder.C_ORDER

    in_axis = 0
    for ind in index:
        if _is_bool_index(ind):
            in_axis += ind.ndim
            return TensorOrder.C_ORDER
        elif _is_fancy_index(ind):
            in_axis += 1
            return TensorOrder.C_ORDER
        elif ind is None:
            continue
        elif isinstance(ind, slice):
            shape = a.shape[in_axis]
            slc = ind.indices(shape)
            if slc[0] == 0 and slc[1] == shape and slc[2] == 1:
                continue
            else:
                return TensorOrder.C_ORDER
        else:
            assert isinstance(ind, Integral)
            in_axis += 1
            return TensorOrder.C_ORDER

    return TensorOrder.F_ORDER


def _getitem_nocheck(a, item, convert_bool_to_fancy=None):
    index = process_index(a.ndim, item,
                          convert_bool_to_fancy=convert_bool_to_fancy)
    if convert_bool_to_fancy is False:
        # come from __setitem__, the bool index is not converted to fancy index
        # if multiple bool indexes or bool + fancy indexes exist,
        # thus the shape will be wrong,
        # here we just convert when calculating shape,
        # refer to issue #1282.
        shape = calc_shape(a.shape, process_index(a.ndim, index))
    else:
        shape = calc_shape(a.shape, index)
    tensor_order = _calc_order(a, index)
    op = TensorIndex(dtype=a.dtype, sparse=a.issparse(), indexes=index,
                     create_view=_is_create_view(index))
    return op(a, index, tuple(shape), order=tensor_order)


def _getitem(a, item):
    if isinstance(item, (list, tuple)) and \
            all(isinstance(it, slice) and it == slice(None) for it in item):
        # nothing to do
        return a

    # TODO(jisheng): field access, e.g. t['a'], t[['a', 'b']]
    return _getitem_nocheck(a, item)
